# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import json
import os
import queue
import time

import confluent_kafka as kafka

import tests.base
from tests.base import (
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)


FIXTURE_DIR = os.path.join(tests.base.FIXTURE_DIR, 'gerrit')


class FakeKafkaMessage:
    def __init__(self, topic, offset, value, error=None):
        self._topic = topic
        self._error = error
        self._offset = offset
        if error:
            self._value = None
        else:
            self._value = value

    def error(self):
        return self._error

    def value(self):
        return self._value

    def partition(self):
        return 0

    def topic(self):
        return self._topic

    def offset(self):
        return self._offset


class FakeKafkaConsumer:
    def __init__(self, config, logger):
        self.config = config
        self.logger = logger
        self.topics = None
        self._queue = queue.Queue()
        self.closed = 0
        self._offset = 0

    def put(self, data):
        self._queue.put(data)

    def subscribe(self, topics):
        self.topics = topics

    def poll(self, timeout=0):
        try:
            data = self._queue.get(timeout=timeout)
            self._queue.task_done()
            if isinstance(data, kafka.KafkaError):
                return FakeKafkaMessage(
                    'gerrit', self._offset, None, error=data)
            self._offset += 1
            return FakeKafkaMessage('gerrit', self._offset - 1, data)
        except queue.Empty:
            return None

    def close(self):
        self.closed += 1


def serialize(event):
    return json.dumps(event).encode('utf8')


class TestGerritEventSourceKafka(ZuulTestCase):
    config_file = 'zuul-gerrit-kafka.conf'

    def setUp(self):
        self.useFixture(fixtures.MonkeyPatch(
            'zuul.driver.gerrit.gerriteventkafka.kafka.Consumer',
            FakeKafkaConsumer))
        super().setUp()

    @simple_layout('layouts/simple.yaml')
    @okay_tracebacks('Broker disconnected before response received')
    def test_kafka(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        consumer = self.fake_gerrit.event_thread.consumer
        self.fake_gerrit.event_thread.RECONNECTION_DELAY = 1

        # Assert we passed the required config entries
        self.assertTrue(isinstance(consumer.config, dict))
        self.assertTrue('bootstrap.servers' in consumer.config)
        self.assertTrue('group.id' in consumer.config)

        # Exercise error handling
        err = kafka.KafkaError(kafka.KafkaError._PARTITION_EOF)
        consumer.put(err)

        # Exercise reconnection
        err = kafka.KafkaError(kafka.KafkaError.NETWORK_EXCEPTION)
        consumer.put(err)

        for _ in iterate_timeout(60, 'wait for reconnect'):
            if consumer is not self.fake_gerrit.event_thread.consumer:
                break
            time.sleep(0.2)

        consumer = self.fake_gerrit.event_thread.consumer
        self.additional_event_queues.append(consumer._queue)

        consumer.put(serialize(A.getPatchsetCreatedEvent(1)))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1')
        ])
        self.assertEqual(A.reported, 1, "A should be reported")

        self.assertTrue(consumer._queue.empty())



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2016 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import threading
import time
import random
from collections import defaultdict
from uuid import uuid4

from apscheduler.schedulers.background import BackgroundScheduler
from opentelemetry import trace

from zuul.driver import Driver, TriggerInterface
from zuul.driver.timer import timertrigger
from zuul.driver.timer import timermodel
from zuul.driver.timer.timermodel import TimerTriggerEvent
from zuul.driver.timer.crontrigger import ZuulCronTrigger
from zuul.lib.logutil import get_annotated_logger
from zuul.zk.election import SessionAwareElection


class TimerDriver(Driver, TriggerInterface):
    name = 'timer'
    election_root = "/zuul/scheduler/timer-election"
    log = logging.getLogger("zuul.TimerDriver")
    tracer = trace.get_tracer("zuul")

    def __init__(self):
        self.log.debug("Starting apscheduler")
        self.apsched = BackgroundScheduler()
        self.apsched.start()
        self.tenant_jobs = {}
        self.election = None
        self.election_thread = None
        self.election_won = False
        # Mapping of locks: canonical project name -> lock
        # The lock are used to avoid concurrent update errors when a
        # lot of periodic pipelines are triggering simultanously.
        self.project_update_locks = defaultdict(threading.Lock)
        self.stop_event = threading.Event()
        self.stopped = False

    def registerScheduler(self, scheduler):
        self.sched = scheduler
        self.election = SessionAwareElection(
            self.sched.zk_client.client, self.election_root)
        self.election_thread = threading.Thread(name="TimerElection",
                                                target=self._runElection,
                                                daemon=True)
        self.election_thread.start()

    def _runElection(self):
        self.log.debug("Starting timer election loop")
        # If there are other contenders, wait until we are primed so
        # that we have all the tenant information.  If there are no
        # other schedulers running, we can start ASAP so we emit
        # events for the early tenants.
        try:
            if self.election.lock.contenders():
                self.log.debug("Waiting for scheduler to be primed")
                self.sched.primed_event.wait()
        except Exception:
            self.log.exception("Error in timer election startup:")
            # Just keep going
        while not self.stopped:
            try:
                self.log.info("Running timer election")
                self.election.run(self._electionInner)
            except Exception:
                self.log.exception("Error in timer election:")
        self.log.debug("Exiting timer election loop")

    def _electionInner(self):
        try:
            self.election_won = True
            self.log.info("Won timer election")
            self.stop_event.wait()
        finally:
            self.election_won = False
            self.stop_event.clear()
            self.log.debug("Timer election tenure ended")

    def reconfigure(self, tenant):
        if self.stopped:
            return
        if not self.apsched:
            # Handle possible reuse of the driver without connection objects.
            self.log.debug("Starting apscheduler on reconfigure")
            self.apsched = BackgroundScheduler()
            self.apsched.start()
        self._addJobs(tenant)

    def _removeJobs(self, tenant, new_jobs):
        # Compare existing jobs to new jobs and remove any that should
        # not be present.
        existing_jobs = self.tenant_jobs.get(tenant.name)
        if not existing_jobs:
            self.log.debug("No jobs to remove for %s, %s remain",
                           tenant, len(new_jobs))
            return
        to_remove = set(existing_jobs.keys()) - set(new_jobs.keys())
        for key in to_remove:
            job = existing_jobs[key]
            job.remove()
        self.log.debug("Removed %s jobs for %s, %s remain",
                       len(to_remove), tenant, len(new_jobs))

    def _addJobs(self, tenant):
        jobs = {}
        for manager in tenant.layout.pipeline_managers.values():
            pipeline = manager.pipeline
            for ef in pipeline.event_filters:
                if not isinstance(ef.trigger, timertrigger.TimerTrigger):
                    continue
                for timespec in ef.timespecs:
                    parts = timespec.split()
                    if len(parts) < 5 or len(parts) > 7:
                        self.log.error(
                            "Unable to parse time value '%s' "
                            "defined in pipeline %s" % (
                                timespec,
                                pipeline.name))
                        continue
                    try:
                        cron_args = dict(
                            minute=parts[0],
                            hour=parts[1],
                            day=parts[2],
                            month=parts[3],
                            day_of_week=parts[4],
                            second=None,
                        )
                        if len(parts) > 5:
                            cron_args['second'] = parts[5]
                        if len(parts) > 6:
                            jitter = int(parts[6])
                        else:
                            jitter = None
                        # Trigger any value errors by creating a
                        # throwaway object.
                        ZuulCronTrigger(jitter=jitter, **cron_args)
                    except ValueError:
                        self.log.error(
                            "Unable to create CronTrigger "
                            "for value '%s' defined in "
                            "pipeline %s",
                            timespec,
                            pipeline.name)
                        continue

                    self._addJobsInner(tenant, pipeline,
                                       cron_args, jitter, timespec,
                                       ef.dereference, jobs)
        self._removeJobs(tenant, jobs)
        self.tenant_jobs[tenant.name] = jobs

    def _addJobsInner(self, tenant, pipeline, cron_args, jitter,
                      timespec, dereference, jobs):
        # jobs is a dict of args->job that we mutate
        existing_jobs = self.tenant_jobs.get(tenant.name, {})
        start_job_count = len(jobs)
        for project_name, pcs in tenant.layout.project_configs.items():
            # timer operates on branch heads and doesn't need
            # speculative layouts to decide if it should be
            # enqueued or not.  So it can be decided on cached
            # data if it needs to run or not.
            pcst = tenant.layout.getAllProjectConfigs(project_name)
            if not [True for pc in pcst if pipeline.name in pc.pipelines]:
                continue

            try:
                for branch in tenant.getProjectBranches(project_name):
                    args = (tenant.name, pipeline.name, project_name,
                            branch, dereference, timespec,)
                    existing_job = existing_jobs.get(args)
                    if jitter:
                        # Resolve jitter here so that it is the same
                        # on every scheduler for a given
                        # project-branch, assuming the same
                        # configuration.
                        prng_init = dict(
                            tenant=tenant.name,
                            project=project_name,
                            branch=branch,
                        )
                        prng_seed = json.dumps(prng_init, sort_keys=True)
                        prng = random.Random(prng_seed)
                        job_jitter = prng.uniform(0, jitter)
                    else:
                        job_jitter = None

                    if existing_job:
                        job = existing_job
                    else:
                        # The 'misfire_grace_time' argument is set to
                        # None to disable checking if the job missed
                        # its run time window.  This ensures we don't
                        # miss a trigger when the job is delayed due
                        # to e.g. high scheduler load. Those short
                        # delays are not a problem for our trigger
                        # use-case.
                        trigger = ZuulCronTrigger(
                            jitter=job_jitter, **cron_args)
                        job = self.apsched.add_job(
                            self._onTrigger, trigger=trigger,
                            args=args, misfire_grace_time=None)
                    jobs[job.args] = job
            except Exception:
                self.log.exception("Unable to create APScheduler job for "
                                   "%s %s %s",
                                   tenant, pipeline, project_name)
        end_job_count = len(jobs)
        self.log.debug("Added %s jobs for %s %s with %s existing tenant jobs",
                       (end_job_count - start_job_count),
                       tenant, pipeline,
                       len(existing_jobs))

    def _onTrigger(self, tenant_name, pipeline_name, project_name, branch,
                   dereference, timespec):
        try:
            if not self.election_won:
                return

            if not self.election.is_still_valid():
                self.log.debug("Noticed election is no longer valid")
                self.stop_event.set()
                return

            attributes = {
                "timespec": timespec,
            }
            with self.tracer.start_as_current_span(
                    "TimerEvent", attributes=attributes):
                self._dispatchEvent(tenant_name, pipeline_name, project_name,
                                    branch, dereference, timespec)
        except Exception:
            self.stop_event.set()
            self.log.exception("Error when dispatching timer event")

    def _dispatchEvent(self, tenant_name, pipeline_name, project_name,
                       branch, dereference, timespec):
        self.log.debug('Got trigger for tenant %s and pipeline %s '
                       'project %s branch %s with timespec %s',
                       tenant_name, pipeline_name, project_name,
                       branch, timespec)
        try:
            tenant = self.sched.abide.tenants[tenant_name]
            (trusted, project) = tenant.getProject(project_name)
            event = TimerTriggerEvent()
            event.type = 'timer'
            event.timespec = timespec
            event.forced_pipeline = pipeline_name
            event.project_hostname = project.canonical_hostname
            event.project_name = project.name
            event.ref = 'refs/heads/%s' % branch
            event.branch = branch
            event.zuul_event_id = str(uuid4().hex)
            event.timestamp = time.time()
            event.arrived_at_scheduler_timestamp = event.timestamp
            # Refresh the branch in order to update the item in the
            # change cache.
            change_key = project.source.getChangeKey(event)
            with self.project_update_locks[project.canonical_name]:
                if dereference:
                    event.newrev = project.source.getProjectBranchSha(
                        project, branch)
                else:
                    event.newrev = None
                project.source.getChange(change_key, refresh=True,
                                         event=event)
            log = get_annotated_logger(self.log, event)
            log.debug("Adding event")
            self.sched.pipeline_trigger_events[tenant.name][
                pipeline_name
            ].put(self.name, event)
        except Exception:
            self.log.exception("Error dispatching timer event for "
                               "tenant %s project %s branch %s",
                               tenant_name, project_name, branch)

    def stop(self):
        self.log.debug("Stopping timer driver")
        self.stopped = True
        self.stop_event.set()
        if self.apsched:
            self.log.debug("Stopping apscheduler")
            self.apsched.shutdown()
            self.apsched = None
            self.log.debug("Stopped apscheduler")
        if self.election:
            self.log.debug("Stopping election")
            self.election.cancel()
            self.log.debug("Stopped election")
        if self.election_thread:
            self.log.debug("Stopping election thread")
            self.election_thread.join()
            self.log.debug("Stopped election thread")
        self.log.debug("Stopped timer driver")

    def getTrigger(self, connection_name, config=None):
        return timertrigger.TimerTrigger(self, config)

    def getTriggerSchema(self):
        return timertrigger.getSchema()

    def getTriggerEventClass(self):
        return timermodel.TimerTriggerEvent



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This file contains provider-related schema chunks that can be reused
# by multiple drivers.  When adding new configuration options, if they
# can be used by more than one driver, add them here instead of in the
# driver.

import voluptuous as vs
from zuul.lib.voluputil import (
    Constant,
    Nullable,
    Optional,
    Required,
    assemble,
)

# Labels

# The label attributes which can appear either in the main body of the
# section stanza, or in a section/provider label, or in a standalone
# label.
common_label = vs.Schema({
    Optional(
        'boot-timeout', default=300,
        doc="""The time (in seconds) to wait for a node to boot.""",
    ): int,
    Optional(
        'max-ready-age', default=0,
        doc="""\
        The time (in seconds) an unassigned node should stay in ready state.
        """
    ): int,
    Optional(
        'max-age', default=0,
        doc="""\
        The time (in seconds) since creation that a node may be
        available for use.  Ready nodes older than this time will be
        deleted.
        """
    ): int,
    Optional(
        'min-retention-time', default=0,
        doc="""\
        The time (in seconds) since an instance was launched, during
        which a node will not be deleted. For node resources with
        minimum billing times, this can be used to ensure that the
        instance is retained for at least the minimum billing interval.

        This setting takes precedence over `max-[ready-]age`.
        """
    ): int,
    Optional(
        'snapshot-timeout', default=3600,
        doc="""The time (in seconds) to wait for a snapshot to complete."""
    ): int,
    Optional(
        'snapshot-expiration', default=3600 * 24 * 7,
        doc="""The time (in seconds) until a snapshot expires."""
    ): int,
    Optional(
        'slots', default=1,
        doc="""\
        How many jobs are permitted run on the same node simultaneously."""
    ): int,
    Optional(
        'reuse', default=False,
        doc="""\
        Should the node be reused (True) or deleted (False) after use."""
    ): bool,
    Optional(
        'executor-zone',
        doc="""\
        Specify that a Zuul executor in the specified zone is
        used to run jobs with nodes from this label.
        """,
    ): Nullable(str),
    Optional(
        'tags',
        doc="""\
        A dictionary of tags to add to nodes.  Avoid the use of
        `zuul_` as a key prefix since Zuul uses this for internal
        values.
        """,
        default=dict): {str: str},
    Optional(
        'final',
        doc="""\
        Whether the configuration of the label may be updated
        by values in label-defaults or overidden with a new definition
        by sections or providers lower in the hierarchy than the point
        at which the final attribute is applied.""",
        default=False): vs.Any(
            Constant(True,
                     doc="The label may not be updated or overidden."),
            Constant(False,
                     doc="The label may be updated or overidden."),
            Constant('allow-override',
                     doc="""\
                     The label may not be updated by label-defaults
                     but may be explicitly overidden by redefining
                     it in a new 'label' entry.""")),
})

# The label attributes that can appear in a section/provider label or
# a standalone label (but not in the section body).
base_label = vs.Schema({
    Required('project_canonical_name'): str,
    Required('config_hash'): str,
    Required(
        'name',
        doc="""\
        The name of the label.  Used to refer to the label in Zuul
        configuration.""",
    ): str,
    Optional(
        'description',
        doc="""\
        A textual description of the label for reference purposes.""",
    ): Nullable(str),
    Optional(
        'image',
        doc="""\
        The image to use with this label.
        """,
    ): Nullable(str),
    Optional(
        'flavor',
        doc="""\
        The flavor to use with this label.
        """,
    ): Nullable(str),
    Optional(
        'min-ready',
        doc="""\
        Minimum number of instances that should be in a ready
        state. Zuul always creates more nodes as necessary in
        response to demand, but setting ``min-ready`` can speed
        processing by attempting to keep nodes on-hand and ready for
        immedate use.  This is best-effort based on available
        capacity and is not a guaranteed allocation.  The default of 0
        means that Zuul will only create nodes of this label when
        there is demand.
        """,
        default=0): int,
})

# Azure doesn't take a key-name, so this is separate.
host_key_checking = vs.Schema({
    Optional(
        'host-key-checking',
        doc="""\
        Whether to validate SSH host keys.  When true, this helps
        ensure that nodes are ready to receive SSH connections before
        they are used.  When set to false, Zuul will not attempt to
        ssh-keyscan nodes after they are booted.  Disable this if the
        zuul-launcher and the nodes it launches are on different
        networks, where the launcher is unable to reach the nodes
        directly.""",
        default=True,
    ): bool,
})

# Label attributes that are common to any kind of ssh-based driver.
ssh_label = assemble(
    vs.Schema({
        Optional(
            'key-name',
            doc="""\
            The name of a keypair that will be used when
            booting the node.
            """,
        ): Nullable(str),
    }),
    host_key_checking,
)

# Images

# The image attributes which can appear either in the main body of the
# section stanza, or in a section/provider image, or in a standalone
# image.
common_image = vs.Schema({
    Optional(
        'username',
        doc="""\
        The username Zuul should use when connecting to the node.
        """,
    ): Nullable(str),
    Optional(
        'connection-type',
        doc="""\
        The connection type that a consumer should use when connecting
        to the node.
        """,
    ): Nullable(vs.Any(
        Constant('winrm',
                 doc="A winrm connection."),
        Constant('ssh',
                 doc="An ssh connection."),
    )),
    Optional(
        'connection-port',
        doc="""\
        The port that Zuul should use when connecting to the node.
        For most nodes this is not necessary. This defaults to 22 when
        ``connection-type`` is 'ssh' and 5986 when it is 'winrm'.""",
    ): Nullable(int),
    Optional(
        'python-path',
        doc="""\
        The path of the default python interpreter.  Used by Zuul to set
        ``ansible_python_interpreter``.  The special value ``auto`` will
        direct Zuul to use inbuilt Ansible logic to select the
        interpreter.""",
    ): Nullable(str),
    Optional(
        'shell-type',
        doc="""\
        The shell type of the node's default shell executable. Used by Zuul
        to set ``ansible_shell_type``. This setting should only be used

        - For a windows image with the experimental `connection-type` ``ssh``
          in which case ``cmd`` or ``powershell`` should be set
          and reflect the node's ``DefaultShell`` configuration.
        - If the default shell is not Bourne compatible (sh), but instead
          e.g. ``csh`` or ``fish``, and the user is aware that there is a
          long-standing issue with ``ansible_shell_type`` in combination
          with ``become``.
        """,
    ): Nullable(str),
    Optional(
        'import-timeout',
        doc="""\
        The limit on the amount of time a successful image import can
        take.""",
        default=300): int,
    Optional(
        'final',
        doc="""\
        Whether the configuration of the label may be updated
        by values in label-defaults or overidden with a new definition
        by sections or providers lower in the hierarchy than the point
        at which the final attribute is applied.""",
        default=False): vs.Any(
            Constant(True,
                     doc="The label may not be updated or overidden."),
            Constant(False,
                     doc="The label may be updated or overidden."),
            Constant('allow-override',
                     doc="""\
                     The label may not be updated by label-defaults
                     but may be explicitly overidden by redefining
                     it in a new 'label' entry.""")),
})

# Same as above, but only for cloud drivers.
cloud_image = vs.Schema({
    Optional(
        'userdata',
        doc="""\
        A string of userdata for a node.  Systems such as "cloud-init"
        may use this to configure the node on boot.
        """,
    ): Nullable(str),
})

# Same as above, but only for zuul images.
common_image_zuul = vs.Schema({
    Optional(
        'upload-methods',
        doc="""\
        An ordered list of methods to use when creating an image in
        the provider.""",
        default=['copy', 'import', 'upload'],
    ): [vs.Any(
        Constant('copy',
                 doc="""\
                 Copy the image from another provider if available.
                 """),
        Constant('import',
                 doc="""\
                 Import the image directly from its storage location.
                 """),
        Constant('upload',
                 doc="""\
                 Download the image from its storage location and
                 upload it to the provider."""),
    )],
    Optional(
        'tags',
        doc="""\
        A dictionary of tags to add to uploaded images, and to nodes
        created from them.  Avoid the use of `zuul_` as a key prefix since
        Zuul uses this for internal values.
        """,
        default=dict,
    ): {str: str},
})

# The image attributes that, in addition to those above, can appear in
# a section/provider image or a standalone image (but not in the
# section body).
base_image = vs.Schema({
    Required('project_canonical_name'): str,
    Required('config_hash'): str,
    Required('branch', doc="UNDOCUMENTED"): str,
    Required(
        'name',
        doc="""\
        The name of the image.  Used to refer to the image in Zuul
        configuration.""",
    ): str,
    Optional(
        'description',
        doc="""\
        A textual description of the image for reference purposes.""",
    ): Nullable(str),
    Required(
        'type',
        doc="""\
        The type of image."""
    ): vs.Any(
        Constant('cloud',
                 doc="An image that already existis in the provider."),
        Constant('zuul',
                 doc="An image that is built and managed by Zuul."),
    ),
})

# Flavors

# The flavor attributes that can appear in a section/provider flavor or
# a standalone flavor (but not in the section body).
base_flavor = vs.Schema({
    Required('project_canonical_name'): str,
    Required('config_hash'): str,
    Required(
        'name',
        doc="""\
        The name of the flavor.  Used to refer to the flavor in Zuul
        configuration.""",
    ): str,
    Optional(
        'description',
        doc="""\
        A textual description of the image for reference purposes.""",
    ): Nullable(str),
})

common_flavor = vs.Schema({
    Optional(
        'final',
        doc="""\
        Whether the configuration of the flavor may be updated
        by values in flavor-defaults or overidden with a new definition
        by sections or providers lower in the hierarchy than the point
        at which the final attribute is applied.""",
        default=False): vs.Any(
            Constant(True,
                     doc="The flavor may not be updated or overidden."),
            Constant(False,
                     doc="The flavor may be updated or overidden."),
            Constant('allow-override',
                     doc="""\
                     The flavor may not be updated by flavor-defaults
                     but may be explicitly overidden by redefining
                     it in a new 'flavor' entry.""")),
})

# Flavor attributes that are common to any kind of cloud driver.
cloud_flavor = vs.Schema({
    Optional('public-ipv4',
             doc="""\
             Whether a public IPv4 address should be attached to nodes.""",
             default=False): bool,
    Optional('public-ipv6',
             doc="""\
             Whether a public IPv6 address should be attached to nodes.""",
             default=False): bool,
})



# Copyright 2012-2014 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import voluptuous as v
from zuul.trigger import BaseTrigger
from zuul.driver.zuul.zuulmodel import ZuulEventFilter
from zuul.driver.util import scalar_or_list, to_list, make_regex, ZUUL_REGEX


class ZuulTrigger(BaseTrigger):
    name = 'zuul'
    log = logging.getLogger("zuul.ZuulTrigger")

    def __init__(self, connection, config=None):
        super(ZuulTrigger, self).__init__(connection, config)
        self._handle_parent_change_enqueued_events = False
        self._handle_project_change_merged_events = False

    def getEventFilters(self, connection_name, trigger_conf,
                        parse_context):
        efilters = []
        pcontext = parse_context
        for trigger in to_list(trigger_conf):
            with pcontext.confAttr(trigger, 'event') as attr:
                types = [make_regex(x, pcontext)
                         for x in to_list(attr)]
            with pcontext.confAttr(trigger, 'pipeline') as attr:
                pipelines = [make_regex(x, pcontext)
                             for x in to_list(attr)]
            f = ZuulEventFilter(
                connection_name=connection_name,
                trigger=self,
                types=types,
                pipelines=pipelines,
                debug=trigger.get('debug'),
            )
            efilters.append(f)

        return efilters


def getSchema():
    zuul_trigger = {
        v.Required('event'):
        scalar_or_list(v.Any(
            'parent-change-enqueued',
            'project-change-merged',
            'image-build',
            'image-validate',
        )),
        'pipeline': scalar_or_list(v.Any(ZUUL_REGEX, str)),
        'debug': bool,
    }

    return zuul_trigger



# Copyright (c) 2016 IBM Corp.
# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import os.path
import re
import select
import threading
import time

from zuul.lib import streamer_utils


class Log(object):

    def __init__(self, path):
        self.path = path
        # The logs are written as binary encoded utf-8, which is what we
        # send over the wire.
        self.file = open(path, 'rb')
        self.stat = os.stat(path)
        self.size = self.stat.st_size


class RequestHandler(streamer_utils.BaseFingerRequestHandler):
    '''
    Class to handle a single log streaming request.

    The log streaming code was blatantly stolen from zuul_console.py. Only
    the (class/method/attribute) names were changed to protect the innocent.
    '''

    log = logging.getLogger("zuul.log_streamer")

    def handle(self):
        try:
            build_uuid = self.getCommand()
        except BrokenPipeError:
            self.log.info("Broken pipe")
            # The stream was disconnected
            return
        except Exception:
            self.log.exception("Failure during getCommand:")
            msg = 'Internal streaming error'
            self.request.sendall(msg.encode("utf-8"))
            return

        # validate build ID
        if not re.match("[0-9A-Fa-f]+$", build_uuid):
            msg = 'Build ID %s is not valid' % build_uuid
            self.request.sendall(msg.encode("utf-8"))
            return

        job_dir = os.path.join(self.server.jobdir_root, build_uuid)
        if not os.path.exists(job_dir):
            msg = 'Build ID %s not found' % build_uuid
            self.request.sendall(msg.encode("utf-8"))
            return

        # check if log file exists
        log_file = os.path.join(job_dir, 'work', 'logs', 'job-output.txt')
        if not os.path.exists(log_file):
            msg = 'Log not found for build ID %s' % build_uuid
            self.request.sendall(msg.encode("utf-8"))
            return

        try:
            self.stream_log(log_file)
        except (ConnectionResetError, BrokenPipeError) as e:
            self.log.error("Streaming failure for build UUID %s: %s",
                           build_uuid, str(e))
        except Exception:
            self.log.exception("Streaming failure for build UUID %s:",
                               build_uuid)
            msg = 'Internal streaming error'
            self.request.sendall(msg.encode("utf-8"))

    def stream_log(self, log_file):
        log = None
        while True:
            if log is not None:
                try:
                    log.file.close()
                except Exception:
                    pass
            while True:
                log = self.chunk_log(log_file)
                if log:
                    break
                time.sleep(0.5)
            while True:
                if self.follow_log(log):
                    break
                else:
                    if log is not None:
                        try:
                            log.file.close()
                        except Exception:
                            pass
                    return

    def chunk_log(self, log_file):
        try:
            log = Log(log_file)
        except Exception:
            return
        while True:
            chunk = log.file.read(4096)
            if not chunk:
                break
            self.request.send(chunk)
        return log

    def follow_log(self, log):
        while True:
            # As long as we have unread data, keep reading/sending
            while True:
                chunk = log.file.read(4096)
                if chunk:
                    self.request.send(chunk)
                else:
                    break

            # See if the file has been removed, meaning we should stop
            # streaming it.
            if not os.path.exists(log.path):
                return False

            # At this point, we are waiting for more data to be written
            time.sleep(0.5)

            # Check to see if the remote end has sent any data, if so,
            # discard
            r, w, e = select.select([self.request], [], [self.request], 0)
            if self.request in e:
                return False
            if self.request in r:
                ret = self.request.recv(1024)
                # Discard anything read, if input is eof, it has
                # disconnected.
                if not ret:
                    return False


class LogStreamerServer(streamer_utils.CustomThreadingTCPServer):

    def __init__(self, *args, **kwargs):
        self.jobdir_root = kwargs.pop('jobdir_root')
        super(LogStreamerServer, self).__init__(*args, **kwargs)


class LogStreamer(object):
    '''
    Class implementing log streaming over the finger daemon port.
    '''

    def __init__(self, host, port, jobdir_root):
        self.log = logging.getLogger('zuul.log_streamer')
        self.log.debug("LogStreamer starting on port %s", port)
        self.server = LogStreamerServer((host, port),
                                        RequestHandler,
                                        jobdir_root=jobdir_root)

        # We start the actual serving within a thread so we can return to
        # the owner.
        self.thd = threading.Thread(target=self._run)
        self.thd.daemon = True
        self.thd.start()

    def _run(self):
        try:
            self.server.serve_forever()
        except Exception:
            self.log.exception("Abnormal termination:")
            raise

    def stop(self):
        if self.thd.is_alive():
            self.server.shutdown()
            self.server.server_close()
            self.thd.join()
            self.log.debug("LogStreamer stopped")



# Copyright 2021 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import string
import threading

import testtools

from zuul import model
from zuul.driver import Driver, TriggerInterface
from zuul.lib.connections import ConnectionRegistry
from zuul.zk import ZooKeeperClient, event_queues, sharding
from zuul.zk.components import (
    ComponentRegistry, COMPONENT_REGISTRY
)

from tests.base import (
    BaseTestCase,
    iterate_timeout,
    ZOOKEEPER_SESSION_TIMEOUT,
)


class EventQueueBaseTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.setupZK()

        self.zk_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.addCleanup(self.zk_client.disconnect)
        self.zk_client.connect()
        self.component_registry = ComponentRegistry(self.zk_client)
        # We don't have any other component to initialize the global
        # registry in these tests, so we do it ourselves.
        COMPONENT_REGISTRY.create(self.zk_client)

        self.connections = ConnectionRegistry()
        self.addCleanup(self.connections.stop)


class DummyEvent(model.AbstractEvent):

    def toDict(self):
        return {}

    def updateFromDict(self):
        pass

    @classmethod
    def fromDict(cls, d):
        return cls()


class DummyEventQueue(event_queues.ZooKeeperEventQueue):

    def put(self, event):
        self._put(event.toDict())

    def __iter__(self):
        for data, ack_ref, _ in self._iterEvents():
            event = DummyEvent.fromDict(data)
            event.ack_ref = ack_ref
            yield event


class TestEventQueue(EventQueueBaseTestCase):

    def setUp(self):
        super().setUp()
        self.queue = DummyEventQueue(self.zk_client, "root")

    def test_missing_ack_ref(self):
        # Every event from a ZK event queue should have an ack_ref
        # attached to it when it is deserialized; ensure that an error
        # is raised if we try to ack an event without one.
        with testtools.ExpectedException(RuntimeError):
            self.queue.ack(DummyEvent())

    def test_double_ack(self):
        # Test that if we ack an event twice, an exception isn't
        # raised.
        self.queue.put(DummyEvent())
        self.assertEqual(len(self.queue), 1)

        event = next(iter(self.queue))
        self.queue.ack(event)
        self.assertEqual(len(self.queue), 0)

        # Should not raise an exception
        self.queue.ack(event)

    def test_invalid_json_ignored(self):
        # Test that invalid json is automatically removed.
        event_path = self.queue._put({})
        self.zk_client.client.set(event_path, b"{ invalid")

        self.assertEqual(len(self.queue), 1)
        self.assertEqual(list(self.queue._iterEvents()), [])
        self.assertEqual(len(self.queue), 0)

    def test_overrun(self):
        # Test that we discard events if we're at the max.
        self.queue.MAX_QUEUE = 1

        self.queue.put(DummyEvent())
        self.assertEqual(len(self.queue), 1)

        self.queue.put(DummyEvent())
        self.assertEqual(len(self.queue), 1)


class DummyTriggerEvent(model.TriggerEvent):
    pass


class DummyDriver(Driver, TriggerInterface):
    name = driver_name = "dummy"

    def getTrigger(self, connection, config=None):
        pass

    def getTriggerSchema(self):
        pass

    def getTriggerEventClass(self):
        return DummyTriggerEvent


class TestTriggerEventQueue(EventQueueBaseTestCase):

    def setUp(self):
        super().setUp()
        self.driver = DummyDriver()
        self.connections.registerDriver(self.driver)

    def test_sharded_tenant_trigger_events(self):
        # Test enqueue/dequeue of the tenant trigger event queue.
        queue = event_queues.TenantTriggerEventQueue(
            self.zk_client, self.connections, "tenant"
        )

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = DummyTriggerEvent()
        data = {'test': ''.join(
            random.choice(string.ascii_letters + string.digits)
            for x in range(sharding.NODE_BYTE_SIZE_LIMIT * 2))}
        event.data = data

        queue.put(self.driver.driver_name, event)
        queue.put(self.driver.driver_name, event)

        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        processed = 0
        for event in queue:
            self.assertIsInstance(event, DummyTriggerEvent)
            processed += 1

        self.assertEqual(processed, 2)
        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        acked = 0
        for event in queue:
            queue.ack(event)
            self.assertEqual(event.data, data)
            acked += 1

        self.assertEqual(acked, 2)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

    def test_lost_sharded_tenant_trigger_events(self):
        # Test cleanup when we write out side-channel data but not an
        # event.
        queue = event_queues.TenantTriggerEventQueue(
            self.zk_client, self.connections, "tenant"
        )

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = DummyTriggerEvent()
        data = {'test': ''.join(
            random.choice(string.ascii_letters + string.digits)
            for x in range(sharding.NODE_BYTE_SIZE_LIMIT * 2))}
        event.data = data

        queue.put(self.driver.driver_name, event)
        queue.put(self.driver.driver_name, event)

        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        # Delete the first event (but not its sharded data)
        events = list(queue)
        self.zk_client.client.delete(events[0].ack_ref.path)
        # There should still be 2 data nodes
        self.assertEqual(len(
            self.zk_client.client.get_children(queue.data_root)), 2)

        # Clean up lost data
        queue.cleanup(age=0)
        # There should only be one data node remaining
        self.assertEqual(len(
            self.zk_client.client.get_children(queue.data_root)), 1)

        # Ack the second event
        queue.ack(events[1])

        # Assert there are no side channel data nodes
        self.assertEqual(len(
            self.zk_client.client.get_children(queue.data_root)), 0)

    def test_tenant_trigger_events(self):
        # Test enqueue/dequeue of the tenant trigger event queue.
        queue = event_queues.TenantTriggerEventQueue(
            self.zk_client, self.connections, "tenant"
        )

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = DummyTriggerEvent()
        queue.put(self.driver.driver_name, event)
        queue.put(self.driver.driver_name, event)

        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        processed = 0
        for event in queue:
            self.assertIsInstance(event, DummyTriggerEvent)
            processed += 1

        self.assertEqual(processed, 2)
        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        acked = 0
        for event in queue:
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 2)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

    def test_pipeline_trigger_events(self):
        # Test enqueue/dequeue of pipeline-specific trigger event
        # queues.
        registry = event_queues.PipelineTriggerEventQueue.createRegistry(
            self.zk_client, self.connections
        )

        queue = registry["tenant"]["pipeline"]
        self.assertIsInstance(queue, event_queues.TriggerEventQueue)

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = DummyTriggerEvent()
        queue.put(self.driver.driver_name, event)

        self.assertEqual(len(queue), 1)
        self.assertTrue(queue.hasEvents())

        other_queue = registry["other_tenant"]["pipeline"]
        self.assertEqual(len(other_queue), 0)
        self.assertFalse(other_queue.hasEvents())

        acked = 0
        for event in queue:
            self.assertIsInstance(event, DummyTriggerEvent)
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 1)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())


class TestManagementEventQueue(EventQueueBaseTestCase):

    def test_management_events(self):
        # Test enqueue/dequeue of the tenant management event queue.
        queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = model.ReconfigureEvent()
        result_future = queue.put(event, needs_result=False)
        self.assertIsNone(result_future)

        result_future = queue.put(event)
        self.assertIsNotNone(result_future)

        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())
        self.assertFalse(result_future.wait(0.1))

        acked = 0
        for event in queue:
            self.assertIsInstance(event, model.ReconfigureEvent)
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 2)
        self.assertTrue(result_future.wait(5))
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

    def test_management_event_error(self):
        # Test that management event errors are reported.
        queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        event = model.ReconfigureEvent()
        result_future = queue.put(event)

        acked = 0
        for event in queue:
            event.traceback = "hello traceback"
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 1)
        with testtools.ExpectedException(RuntimeError, msg="hello traceback"):
            self.assertFalse(result_future.wait(5))

    def test_event_merge(self):
        # Test that similar management events (eg, reconfiguration of
        # two projects) can be merged.
        queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        event = model.TenantReconfigureEvent("tenant", "project", "master")
        queue.put(event, needs_result=False)
        event = model.TenantReconfigureEvent("tenant", "other", "branch")
        queue.put(event, needs_result=False)

        self.assertEqual(len(queue), 2)
        events = list(queue)

        self.assertEqual(len(events), 1)
        event = events[0]
        self.assertEqual(len(event.merged_events), 1)
        self.assertEqual(
            event.project_branches,
            set([("project", "master"), ("other", "branch")])
        )

        queue.ack(event)
        self.assertFalse(queue.hasEvents())

    def test_event_ltime(self):
        tenant_queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        registry = event_queues.PipelineManagementEventQueue.createRegistry(
            self.zk_client
        )

        event = model.ReconfigureEvent()
        tenant_queue.put(event, needs_result=False)
        self.assertTrue(tenant_queue.hasEvents())

        pipeline_queue = registry["tenant"]["pipeline"]
        self.assertIsInstance(
            pipeline_queue, event_queues.ManagementEventQueue
        )
        processed_events = 0
        for event in tenant_queue:
            processed_events += 1
            event_ltime = event.zuul_event_ltime
            self.assertGreater(event_ltime, -1)
            # Forward event to pipeline management event queue
            pipeline_queue.put(event)

        self.assertEqual(processed_events, 1)
        self.assertTrue(pipeline_queue.hasEvents())

        processed_events = 0
        for event in pipeline_queue:
            pipeline_queue.ack(event)
            processed_events += 1
            self.assertEqual(event.zuul_event_ltime, event_ltime)

        self.assertEqual(processed_events, 1)

    def test_pipeline_management_events(self):
        # Test that when a management event is forwarded from the
        # tenant to the a pipeline-specific queue, it is not
        # prematurely acked and the future returns correctly.
        tenant_queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        registry = event_queues.PipelineManagementEventQueue.createRegistry(
            self.zk_client
        )

        event = model.PromoteEvent('tenant', 'check', ['1234,1'])
        result_future = tenant_queue.put(event, needs_result=False)
        self.assertIsNone(result_future)

        result_future = tenant_queue.put(event)
        self.assertIsNotNone(result_future)

        self.assertEqual(len(tenant_queue), 2)
        self.assertTrue(tenant_queue.hasEvents())

        pipeline_queue = registry["tenant"]["pipeline"]
        self.assertIsInstance(
            pipeline_queue, event_queues.ManagementEventQueue
        )
        acked = 0
        for event in tenant_queue:
            self.assertIsInstance(event, model.PromoteEvent)
            # Forward event to pipeline management event queue
            pipeline_queue.put(event)
            tenant_queue.ackWithoutResult(event)
            acked += 1

        self.assertEqual(acked, 2)
        # Event was just forwarded and since we expect a result, the
        # future should not be completed yet.
        self.assertFalse(result_future.wait(0.1))

        self.assertEqual(len(tenant_queue), 0)
        self.assertFalse(tenant_queue.hasEvents())

        self.assertEqual(len(pipeline_queue), 2)
        self.assertTrue(pipeline_queue.hasEvents())

        acked = 0
        for event in pipeline_queue:
            self.assertIsInstance(event, model.PromoteEvent)
            pipeline_queue.ack(event)
            acked += 1

        self.assertEqual(acked, 2)
        self.assertTrue(result_future.wait(5))
        self.assertEqual(len(pipeline_queue), 0)
        self.assertFalse(pipeline_queue.hasEvents())

    def test_management_events_client(self):
        # Test management events from a second client

        queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        # This client will submit a reconfigure event and wait for it.
        external_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.addCleanup(external_client.disconnect)
        external_client.connect()

        external_queue = event_queues.TenantManagementEventQueue(
            external_client, "tenant")

        event = model.ReconfigureEvent()
        result_future = external_queue.put(event)
        self.assertIsNotNone(result_future)

        self.assertEqual(len(queue), 1)
        self.assertTrue(queue.hasEvents())
        self.assertFalse(result_future.wait(0.1))

        acked = 0
        for event in queue:
            self.assertIsInstance(event, model.ReconfigureEvent)
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 1)
        self.assertTrue(result_future.wait(5))
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

    def test_management_events_client_disconnect(self):
        # Test management events from a second client which
        # disconnects before the event is complete.

        queue = event_queues.TenantManagementEventQueue(
            self.zk_client, "tenant")
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        # This client will submit a reconfigure event and disconnect
        # before it's complete.
        external_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.addCleanup(external_client.disconnect)
        external_client.connect()

        external_queue = event_queues.TenantManagementEventQueue(
            external_client, "tenant")

        # Submit the event
        event = model.ReconfigureEvent()
        result_future = external_queue.put(event)
        self.assertIsNotNone(result_future)

        # Make sure the event is in the queue and the result node exists
        self.assertEqual(len(queue), 1)
        self.assertTrue(queue.hasEvents())
        self.assertFalse(result_future.wait(0.1))
        self.assertEqual(len(
            self.zk_client.client.get_children('/zuul/results/management')), 1)

        # Disconnect the originating client
        external_client.disconnect()
        # Ensure the result node is gone
        self.assertEqual(len(
            self.zk_client.client.get_children('/zuul/results/management')), 0)

        # Process the event
        acked = 0
        for event in queue:
            self.assertIsInstance(event, model.ReconfigureEvent)
            queue.ack(event)
            acked += 1

        # Make sure the event has been processed and we didn't
        # re-create the result node.
        self.assertEqual(acked, 1)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())
        self.assertEqual(len(
            self.zk_client.client.get_children('/zuul/results/management')), 0)


class TestResultEventQueue(EventQueueBaseTestCase):

    def test_pipeline_result_events(self):
        # Test enqueue/dequeue of result events.
        registry = event_queues.PipelineResultEventQueue.createRegistry(
            self.zk_client
        )

        queue = registry["tenant"]["pipeline"]
        self.assertIsInstance(queue, event_queues.PipelineResultEventQueue)

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        event = model.BuildStartedEvent(
            "build", "buildset", "job", "job_uuid", "build_request_path", {})
        queue.put(event)

        self.assertEqual(len(queue), 1)
        self.assertTrue(queue.hasEvents())

        other_queue = registry["other_tenant"]["pipeline"]
        self.assertEqual(len(other_queue), 0)
        self.assertFalse(other_queue.hasEvents())

        acked = 0
        for event in queue:
            self.assertIsInstance(event, model.BuildStartedEvent)
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 1)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())


class TestEventWatchers(EventQueueBaseTestCase):

    def setUp(self):
        super().setUp()
        self.driver = DummyDriver()
        self.connections.registerDriver(self.driver)

    def _wait_for_event(self, event):
        for _ in iterate_timeout(5, "event set"):
            if event.is_set():
                break

    def test_tenant_event_watcher(self):
        event = threading.Event()
        event_queues.EventWatcher(self.zk_client, event.set)

        management_queue = (
            event_queues.TenantManagementEventQueue.createRegistry(
                self.zk_client)
        )
        trigger_queue = event_queues.TenantTriggerEventQueue.createRegistry(
            self.zk_client, self.connections
        )
        self.assertFalse(event.is_set())

        management_queue["tenant"].put(model.ReconfigureEvent(),
                                       needs_result=False)
        self._wait_for_event(event)
        event.clear()

        trigger_queue["tenant"].put(self.driver.driver_name,
                                    DummyTriggerEvent())
        self._wait_for_event(event)

    def test_pipeline_event_watcher(self):
        event = threading.Event()
        event_queues.EventWatcher(self.zk_client, event.set)

        management_queues = (
            event_queues.PipelineManagementEventQueue.createRegistry(
                self.zk_client
            )
        )
        trigger_queues = event_queues.PipelineTriggerEventQueue.createRegistry(
            self.zk_client, self.connections
        )
        result_queues = event_queues.PipelineResultEventQueue.createRegistry(
            self.zk_client
        )
        self.assertFalse(event.is_set())

        management_queues["tenant"]["check"].put(model.ReconfigureEvent())
        self._wait_for_event(event)
        event.clear()

        trigger_queues["tenant"]["gate"].put(self.driver.driver_name,
                                             DummyTriggerEvent())
        self._wait_for_event(event)
        event.clear()

        result_event = model.BuildStartedEvent(
            "build", "buildset", "job", "job_uuid", "build_request_path", {})
        result_queues["other-tenant"]["post"].put(result_event)
        self._wait_for_event(event)

    def test_pipeline_event_watcher_recreate(self):
        event = threading.Event()
        watcher = event_queues.EventWatcher(self.zk_client, event.set)

        management_queues = (
            event_queues.PipelineManagementEventQueue.createRegistry(
                self.zk_client
            )
        )
        self.assertFalse(event.is_set())

        management_queues["tenant"]["check"].put(model.ReconfigureEvent())
        self._wait_for_event(event)

        # Wait for the watch to be fully established to avoid race
        # conditions, since the event watcher will also ensure that the
        # trigger and result event paths exist.
        for _ in iterate_timeout(5, "all watches to be established"):
            if watcher.watched_pipelines:
                break

        self.zk_client.client.delete(
            event_queues.PIPELINE_NAME_ROOT.format(
                tenant="tenant", pipeline="check"), recursive=True)
        event.clear()

        management_queues["tenant"]["check"].initialize()
        management_queues["tenant"]["check"].put(model.ReconfigureEvent())
        self._wait_for_event(event)


class TestConnectionEventQueue(EventQueueBaseTestCase):

    def test_connection_events(self):
        # Test enqueue/dequeue of the connection event queue.
        queue = event_queues.ConnectionEventQueue(
            self.zk_client, "dummy", None)

        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        payload = {"message": "hello world!"}
        queue.put(payload)
        queue.put(payload)

        self.assertEqual(len(queue), 2)
        self.assertTrue(queue.hasEvents())

        acked = 0
        for event in queue:
            self.assertIsInstance(event, model.ConnectionEvent)
            self.assertEqual(event, payload)
            queue.ack(event)
            acked += 1

        self.assertEqual(acked, 2)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

    def test_event_watch(self):
        # Test the registered function is called on new events.
        queue = event_queues.ConnectionEventQueue(
            self.zk_client, "dummy", None)

        event = threading.Event()
        queue.registerEventWatch(event.set)
        self.assertFalse(event.is_set())

        queue.put({"message": "hello world!"})
        for _ in iterate_timeout(5, "event set"):
            if event.is_set():
                break

    def test_event_offset(self):
        queue = event_queues.ConnectionEventQueue(
            self.zk_client, "dummy", None)
        self.assertEqual(len(queue), 0)
        self.assertFalse(queue.hasEvents())

        payload = {"message": "hello world!"}
        queue.put(payload)
        queue.put(payload)

        events = list(queue)
        self.assertEqual(len(events), 2)

        skipped_event = events[0]
        offset = queue.eventIdFromAckRef(skipped_event.ack_ref)
        for event in queue.iter(offset):
            self.assertNotEqual(event.ack_ref, skipped_event.ack_ref)
            queue.ack(event)

        self.assertEqual(len(queue), 1)
        for event in queue.iter():
            self.assertEqual(event.ack_ref, skipped_event.ack_ref)
            queue.ack(event)
        self.assertEqual(len(queue), 0)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.driver import Driver, ConnectionInterface, TriggerInterface
from zuul.driver import SourceInterface, ReporterInterface
from zuul.driver.gitlab import gitlabconnection
from zuul.driver.gitlab import gitlabmodel
from zuul.driver.gitlab import gitlabsource
from zuul.driver.gitlab import gitlabreporter
from zuul.driver.gitlab import gitlabtrigger


class GitlabDriver(Driver, ConnectionInterface, TriggerInterface,
                   SourceInterface, ReporterInterface):
    name = 'gitlab'

    def getConnection(self, name, config):
        return gitlabconnection.GitlabConnection(self, name, config)

    def getTrigger(self, connection, config=None):
        return gitlabtrigger.GitlabTrigger(self, connection, config)

    def getTriggerEventClass(self):
        return gitlabmodel.GitlabTriggerEvent

    def getSource(self, connection):
        return gitlabsource.GitlabSource(self, connection)

    def getReporter(self, connection, pipeline, config=None,
                    parse_context=None):
        return gitlabreporter.GitlabReporter(
            self, connection, pipeline, config)

    def getTriggerSchema(self):
        return gitlabtrigger.getSchema()

    def getReporterSchema(self):
        return gitlabreporter.getSchema()

    def getRequireSchema(self):
        return gitlabsource.getRequireSchema()

    def getRejectSchema(self):
        return gitlabsource.getRejectSchema()



# Copyright 2024 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import threading
import time

from collections import defaultdict
from zuul.model import (
    NodesetInfo,
    NodesetRequest,
    ProviderNode,
    STATE_HOLD,
    STATE_IN_USE,
    STATE_READY,
    STATE_USED,
)
from zuul.lib import tracing
from zuul.lib.logutil import get_annotated_logger
from zuul.zk.vendor.watchers import ExistingDataWatch
from zuul.zk.zkobject import ZKContext

from kazoo.exceptions import NoNodeError
from opentelemetry import trace


class LauncherClient:
    log = logging.getLogger("zuul.LauncherClient")
    tracer = trace.get_tracer("zuul")
    # The kind of resources we report stats on.  We need a complete
    # list in order to report 0 level gauges.
    resource_types = ('ram', 'cores', 'instances')
    HOLD_REQUEST_ROOT = '/zuul/hold-requests'

    def __init__(self, zk_client, stop_event, component_info=None):
        self.component_info = component_info
        self.zk_client = zk_client
        self.stop_event = stop_event

    def requestNodeset(self, item, job, priority, relative_priority,
                       preferred_provider):
        log = get_annotated_logger(self.log, item.event)
        labels = [n.label for n in job.nodeset.getNodes()]

        buildset = item.current_build_set
        parent_span = tracing.restoreSpan(buildset.span_info)
        request_time = time.time()
        with trace.use_span(parent_span):
            request_span = self.tracer.start_span(
                "NodesetRequest", start_time=request_time)
        span_info = tracing.getSpanInfo(request_span)

        image_names = getattr(item.event, "image_names", None)
        image_upload_uuid = getattr(item.event, "image_upload_uuid", None)

        with self.createZKContext(None, self.log) as ctx:
            request = NodesetRequest.new(
                ctx,
                **self._getInitialRequestStates(job),
                tenant_name=item.manager.tenant.name,
                pipeline_name=item.manager.pipeline.name,
                buildset_uuid=buildset.uuid,
                job_uuid=job.uuid,
                job_name=job.name,
                labels=labels,
                priority=priority,
                _relative_priority=relative_priority,
                request_time=request_time,
                zuul_event_id=item.event.zuul_event_id,
                span_info=span_info,
                image_names=image_names,
                image_upload_uuid=image_upload_uuid,
                preferred_provider=preferred_provider,
            )
            log.info("Submitted nodeset request %s", request)
        return request

    def reviseRequest(self, request, relative_priority):
        log = get_annotated_logger(self.log, request)
        try:
            with self.createZKContext(None, self.log) as ctx:
                request.revise(ctx, relative_priority=relative_priority)
            log.info("Revised nodeset request %s; relative_priority=%s",
                     request, relative_priority)
        except NoNodeError:
            pass

    def getRequest(self, request_id):
        try:
            with self.createZKContext(None, self.log) as ctx:
                return NodesetRequest.fromZK(ctx, path=None, uuid=request_id)
        except NoNodeError:
            return None

    def deleteRequest(self, request):
        try:
            with self.createZKContext(None, self.log) as ctx:
                request.delete(ctx)
        except NoNodeError:
            pass

    def getRequestIds(self):
        # Do not use this if a cache is available; this should only be
        # used by components (like the scheduler) which do not
        # otherwise have a cache.
        path = f'{NodesetRequest.ROOT}/{NodesetRequest.REQUESTS_PATH}'
        try:
            return self.zk_client.client.get_children(path)
        except NoNodeError:
            return []

    def getProviderNode(self, node_id):
        # Do not use this if a cache is available; this should only be
        # used by components (like the scheduler) which do not
        # otherwise have a cache.
        path = ProviderNode._getPath(node_id)
        with self.createZKContext(None, self.log) as ctx:
            return ProviderNode.fromZK(ctx, path)

    def getNodesetInfo(self, request):
        # TODO: populated other nodeset info fields
        zone = request.executor_zones[0] if request.executor_zones else None
        return NodesetInfo(
            nodes=list(request.nodes),
            zone=zone)

    def acceptNodeset(self, request, nodeset):
        log = get_annotated_logger(self.log, request)
        log.debug("Accepting nodeset for request %s", request)
        try:
            with self.createZKContext(None, self.log) as ctx:
                for node_id, node in zip(
                    request.nodes, nodeset.getNodes()
                ):
                    provider_node = ProviderNode.fromZK(
                        ctx, path=ProviderNode._getPath(node_id))
                    node.updateFromDict({
                        "state": STATE_READY,
                        **provider_node.getNodeData()
                    })
                    node._provider_node = provider_node
                    log.debug("Locking node %s", provider_node)
                    if not provider_node.acquireLock(ctx):
                        raise Exception(f"Failed to lock node {provider_node}")
        except Exception:
            self.returnNodeset(nodeset)
            raise
        finally:
            self.deleteRequest(request)

    def useNodeset(self, nodeset, zuul_event_id=None):
        log = get_annotated_logger(self.log, zuul_event_id)
        log.info("Setting nodeset %s in use", nodeset)
        for node in nodeset.getNodes():
            provider_node = getattr(node, "_provider_node", None)
            if not provider_node:
                raise Exception("No associated provider node for %s", node)
            if not provider_node.hasLock():
                raise Exception("Provider node %s is not locked",
                                provider_node)
            node.state = STATE_IN_USE
            with self.createZKContext(provider_node._lock, log) as ctx:
                provider_node.updateAttributes(
                    ctx, state=ProviderNode.State.IN_USE)

    def returnNodeset(self, nodeset, zuul_event_id=None):
        log = get_annotated_logger(self.log, zuul_event_id)
        log.debug("Returning nodeset %s", nodeset)
        for node in nodeset.getNodes():
            provider_node = getattr(node, "_provider_node", None)
            if not provider_node:
                continue
            with self.createZKContext(provider_node._lock, log) as ctx:
                try:
                    if provider_node.state == provider_node.State.IN_USE:
                        node.state = STATE_USED
                        provider_node.updateAttributes(
                            ctx, state=ProviderNode.State.USED)
                        log.debug("Released %s", provider_node)
                except Exception:
                    log.exception("Unable to return node %s", provider_node)
                finally:
                    try:
                        provider_node.releaseLock(ctx)
                    except Exception:
                        log.exception("Error unlocking node %s", provider_node)

    def snapshotNodeset(self, nodeset, node_ids, zuul_event_id=None):
        log = get_annotated_logger(self.log, zuul_event_id)
        log.debug("Snapshotting nodes %s in nodeset %s", node_ids, nodeset)

        wait_event = threading.Event()
        wait_snapshots = []

        def node_watcher(self, data, zstat, event=None):
            wait_event.set()

        for node in nodeset.getNodes():
            provider_node = getattr(node, "_provider_node", None)
            if not provider_node:
                continue
            if provider_node.uuid not in node_ids:
                continue
            with self.createZKContext(provider_node._lock, log) as ctx:
                try:
                    if provider_node.state == provider_node.State.IN_USE:
                        with provider_node.activeContext(ctx):
                            provider_node.createSnapshot(ctx)
                            provider_node.setState(ProviderNode.State.SNAPSHOT)
                        ExistingDataWatch(self.zk_client.client,
                                          provider_node.snapshot.getPath(),
                                          node_watcher)
                        wait_snapshots.append(provider_node.snapshot)
                        log.debug("Set %s to snapshot", provider_node)
                except Exception:
                    log.exception("Unable to snapshot node %s", provider_node)

        if not wait_snapshots:
            log.debug("Nothing to snapshot in nodeset %s", nodeset)
            return
        done = False
        log.debug("Waiting for snapshots in nodeset %s", nodeset)
        while not done:
            wait_event.wait()
            wait_event.clear()
            done = True
            with self.createZKContext(provider_node._lock, log) as ctx:
                for snapshot in wait_snapshots:
                    snapshot.refresh(ctx)
                    if not snapshot.complete:
                        done = False
        log.debug("Finished snapshots in nodeset %s", nodeset)

    def addResources(self, target, source):
        for key, value in source.items():
            if key in self.resource_types:
                target[key] += value

    def holdNodeSet(self, zk_nodepool, nodeset, request, build, duration,
                    zuul_event_id=None):
        '''
        Perform a hold on the given set of nodes.

        :param NodeSet nodeset: The object containing the set of nodes to hold.
        :param HoldRequest request: Hold request associated with the NodeSet
        '''
        log = get_annotated_logger(self.log, zuul_event_id)
        log.info("Holding nodeset %s", nodeset)
        resources = defaultdict(int)
        nodes = nodeset.getNodes()

        log.info(
            "Nodeset %s with %s nodes was in use for %s seconds for build %s "
            "for project %s",
            nodeset, len(nodeset.nodes), duration, build, request.project)

        node_ids = []
        for node in nodes:
            provider_node = getattr(node, "_provider_node", None)
            if not provider_node:
                raise Exception("No associated provider node for %s", node)
            if not provider_node.hasLock():
                raise Exception("Provider node %s is not locked",
                                provider_node)
            if node.resources:
                self.addResources(resources, node.resources)

            node_ids.append(provider_node.uuid)
            with self.createZKContext(provider_node._lock, log) as ctx:
                with provider_node.activeContext(ctx):
                    provider_node.setState(ProviderNode.State.HOLD)
                    provider_node.comment = request.reason
                    if request.node_expiration:
                        provider_node.hold_expiration = request.node_expiration
                log.debug("Held %s", provider_node)
            node.state = STATE_HOLD

        request.nodes.append(dict(
            niz=True,
            build=build.uuid,
            nodes=node_ids,
        ))
        request.current_count += 1

        # Request has been used at least the maximum number of times so set
        # the expiration time so that it can be auto-deleted.
        if request.current_count >= request.max_count and not request.expired:
            request.expired = time.time()

        # Give ourselves a few seconds to try to obtain the lock rather than
        # immediately give up.
        zk_nodepool.lockHoldRequest(request, timeout=5)

        try:
            zk_nodepool.storeHoldRequest(request)
        except Exception:
            # If we fail to update the request count, we won't consider it
            # a real autohold error by passing the exception up. It will
            # just get used more than the original count specified.
            # It's possible to leak some held nodes, though, which would
            # require manual node deletes.
            log.exception("Unable to update hold request %s:", request)
        finally:
            # Although any exceptions thrown here are handled higher up in
            # _doBuildCompletedEvent, we always want to try to unlock it.
            zk_nodepool.unlockHoldRequest(request)

        # TODO: implement stats reporting
        # if resources and duration:
        #     self.emitStatsResourceCounters(
        #         request.tenant, request.project, resources, duration)

    def markHeldNodesAsUsed(self, hold_request, node_group):
        """
        Changes the state for each held node for the hold request to 'used'.

        :returns: True if all nodes marked USED, False otherwise.
        """

        failure = False
        for node_id in node_group['nodes']:
            provider_node = self.getProviderNode(node_id)
            if not provider_node:
                continue
            if provider_node.state == provider_node.State.USED:
                continue
            with self.createZKContext(None, self.log) as ctx:
                if not provider_node.acquireLock(ctx):
                    raise Exception(f"Failed to lock node {provider_node}")
            with self.createZKContext(provider_node._lock, self.log) as ctx:
                try:
                    with provider_node.activeContext(ctx):
                        provider_node.setState(ProviderNode.State.USED)
                    self.log.debug("Released %s", provider_node)
                except Exception:
                    failure = True
                    self.log.exception("Unable to return node %s",
                                       provider_node)
                finally:
                    try:
                        provider_node.releaseLock(ctx)
                    except Exception:
                        self.log.exception("Error unlocking node %s",
                                           provider_node)

        return not failure

    def createZKContext(self, lock, log):
        if self.component_info:
            identifier = self.component_info.hostname
        else:
            identifier = None
        return ZKContext(self.zk_client, lock, self.stop_event, log,
                         default_lock_identifier=identifier)

    def _getInitialRequestStates(self, job):
        if not job.nodeset.nodes:
            # Nodeless job
            return dict(
                state=NodesetRequest.State.FULFILLED,
                event_state=NodesetRequest.EventState.COMPLETE,
            )
        # Use default states
        return dict()



# Copyright 2021 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from kazoo.exceptions import NoNodeError

from zuul.lib.collections import DefaultKeyDict
from zuul.model import BuildRequest
from zuul.zk.job_request_queue import JobRequestQueue


class ExecutorQueue(JobRequestQueue):
    log = logging.getLogger("zuul.ExecutorQueue")
    request_class = BuildRequest

    def __init__(self, client, root,
                 initial_state_getter,
                 use_cache=True,
                 request_callback=None,
                 event_callback=None):
        self.log.debug("Creating executor queue at root %s", root)
        self._initial_state_getter = initial_state_getter
        super().__init__(
            client, root, use_cache, request_callback, event_callback)

    @property
    def initial_state(self):
        # This supports holding requests in tests
        return self._initial_state_getter()

    def lostRequests(self):
        return super().lostRequests(self.request_class.RUNNING,
                                    self.request_class.PAUSED)


class ExecutorApi:
    log = logging.getLogger("zuul.ExecutorApi")

    def __init__(self, client, zone_filter=None, use_cache=True,
                 build_request_callback=None,
                 build_event_callback=None):
        self.client = client
        self.use_cache = use_cache
        self.request_callback = build_request_callback
        self.event_callback = build_event_callback
        self.zone_filter = zone_filter
        self._watched_zones = set()
        self.root = '/zuul/executor'
        self.unzoned_root = f"{self.root}/unzoned"
        self.zones_root = f"{self.root}/zones"

        self.zone_queues = DefaultKeyDict(
            lambda zone: ExecutorQueue(
                self.client,
                self._getZoneRoot(zone),
                self._getInitialState,
                self.use_cache,
                self.request_callback,
                self.event_callback))

        if zone_filter is None:
            self.registerAllZones()
        else:
            for zone in zone_filter:
                # For the side effect of creating a queue
                self.zone_queues[zone]

    def stop(self):
        for queue in self.zone_queues.values():
            queue.cache.stop()

    def _getInitialState(self):
        return BuildRequest.REQUESTED

    def _getZoneRoot(self, zone):
        if zone is None:
            return self.unzoned_root
        else:
            return f"{self.zones_root}/{zone}"

    def registerAllZones(self):
        # Register a child watch that listens to new zones and automatically
        # registers to them.
        def watch_zones(children):
            for zone in children:
                # For the side effect of creating a queue
                self.zone_queues[zone]

        self.client.client.ChildrenWatch(self.zones_root, watch_zones)
        # For the side effect of creating a queue
        self.zone_queues[None]

    def _getAllZones(self):
        # Get a list of all zones without using the cache.
        try:
            # Get all available zones from ZooKeeper
            zones = self.client.client.get_children(self.zones_root)
            zones.append(None)
        except NoNodeError:
            zones = [None]
        return zones

    # Override JobRequestQueue methods to accomodate the zone dict.

    def inState(self, *states):
        requests = []
        for queue in self.zone_queues.values():
            requests.extend(queue.inState(*states))
        return sorted(requests)

    def next(self):
        for request in self.inState(BuildRequest.REQUESTED):
            for queue in self.zone_queues.values():
                request2 = queue.getRequest(request.uuid)
                if (request2 and
                    request2.state == BuildRequest.REQUESTED and
                    not request2.is_locked):
                    yield request2
                    break

    def submit(self, request, params):
        return self.zone_queues[request.zone].submit(request, params)

    def getRequestUpdater(self, request):
        return self.zone_queues[request.zone].getRequestUpdater(request)

    def update(self, request):
        return self.zone_queues[request.zone].update(request)

    def reportResult(self, request, result):
        return self.zone_queues[request.zone].reportResult(request)

    def get(self, path):
        if path.startswith(self.zones_root):
            # Remove zone root so we end up with: <zone>/requests/<uuid>
            rel_path = path[len(f"{self.zones_root}/"):]
            zone, _, request_uuid = rel_path.split("/")
        else:
            rel_path = path[len(f"{self.unzoned_root}/"):]
            zone = None
            _, request_uuid = rel_path.split("/")
        return self.zone_queues[zone].getRequest(request_uuid)

    def getRequest(self, uuid):
        """Find a build request by its UUID.

        This method will search for the UUID in all available zones.
        """
        for zone in self._getAllZones():
            request = self.zone_queues[zone].getRequest(uuid)
            if request:
                return request
        return None

    def remove(self, request):
        return self.zone_queues[request.zone].remove(request)

    def requestResume(self, request):
        return self.zone_queues[request.zone].requestResume(request)

    def requestCancel(self, request):
        return self.zone_queues[request.zone].requestCancel(request)

    def fulfillResume(self, request):
        return self.zone_queues[request.zone].fulfillResume(request)

    def fulfillCancel(self, request):
        return self.zone_queues[request.zone].fulfillCancel(request)

    def lock(self, request, *args, **kw):
        return self.zone_queues[request.zone].lock(request, *args, **kw)

    def unlock(self, request):
        return self.zone_queues[request.zone].unlock(request)

    def isLocked(self, request):
        return self.zone_queues[request.zone].isLocked(request)

    def lostRequests(self):
        for queue in self.zone_queues.values():
            yield from queue.lostRequests()

    def cleanup(self, age=300):
        for queue in self.zone_queues.values():
            queue.cleanup(age)

    def clearParams(self, request):
        return self.zone_queues[request.zone].clearParams(request)

    def getParams(self, request):
        return self.zone_queues[request.zone].getParams(request)

    def _getAllRequestIds(self):
        ret = []
        for queue in self.zone_queues.values():
            ret.extend(queue._getAllRequestIds())
        return ret

    def waitForSync(self):
        for queue in self.zone_queues.values():
            queue.waitForSync()



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from tests.base import AnsibleZuulTestCase


class TestOpenStack(AnsibleZuulTestCase):
    # A temporary class to experiment with how openstack can use
    # Zuulv3

    tenant_config_file = 'config/openstack/main.yaml'

    def test_nova_master(self):
        A = self.fake_gerrit.addFakeChange('openstack/nova', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('python27').result,
                         'SUCCESS')
        self.assertEqual(self.getJobFromHistory('python35').result,
                         'SUCCESS')
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2,
                         "A should report start and success")
        self.assertEqual(self.getJobFromHistory('python27').node,
                         'ubuntu-xenial')

    def test_nova_mitaka(self):
        self.create_branch('openstack/nova', 'stable/mitaka')
        A = self.fake_gerrit.addFakeChange('openstack/nova',
                                           'stable/mitaka', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('python27').result,
                         'SUCCESS')
        self.assertEqual(self.getJobFromHistory('python35').result,
                         'SUCCESS')
        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2,
                         "A should report start and success")
        self.assertEqual(self.getJobFromHistory('python27').node,
                         'ubuntu-trusty')

    def test_dsvm_keystone_repo(self):
        self.executor_server.keep_jobdir = True
        A = self.fake_gerrit.addFakeChange('openstack/nova', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='dsvm', result='SUCCESS', changes='1,1')])
        build = self.getJobFromHistory('dsvm')

        # Check that a change to nova triggered a keystone clone
        executor_git_dir = os.path.join(self.executor_src_root,
                                        'review.example.com',
                                        'openstack', 'openstack%2Fkeystone',
                                        '.git')
        self.assertTrue(os.path.exists(executor_git_dir),
                        msg='openstack/keystone should be cloned.')

        jobdir_git_dir = os.path.join(build.jobdir.src_root,
                                      'review.example.com',
                                      'openstack', 'keystone', '.git')
        self.assertTrue(os.path.exists(jobdir_git_dir),
                        msg='openstack/keystone should be cloned.')

    def test_dsvm_nova_repo(self):
        self.executor_server.keep_jobdir = True
        A = self.fake_gerrit.addFakeChange('openstack/keystone', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='dsvm', result='SUCCESS', changes='1,1')])
        build = self.getJobFromHistory('dsvm')

        # Check that a change to keystone triggered a nova clone
        executor_git_dir = os.path.join(self.executor_src_root,
                                        'review.example.com',
                                        'openstack', 'openstack%2Fnova',
                                        '.git')
        self.assertTrue(os.path.exists(executor_git_dir),
                        msg='openstack/nova should be cloned.')

        jobdir_git_dir = os.path.join(build.jobdir.src_root,
                                      'review.example.com',
                                      'openstack', 'nova', '.git')
        self.assertTrue(os.path.exists(jobdir_git_dir),
                        msg='openstack/nova should be cloned.')



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import sys


def main():
    console_path = sys.argv[1]
    output_txt_path = sys.argv[2]
    output_json_path = sys.argv[3]
    with open(output_json_path) as f:
        output_json = json.loads(f.read())
    with open(output_txt_path) as f:
        output_txt = f.read()
    with open(console_path) as f:
        console = f.read()

    # Get a list of all the task ids where we skippod log live
    # streaming (but we get one for each host, so use a set to dedup).
    skipped = set()

    for playbook in output_json:
        for play in playbook['plays']:
            for task in play['tasks']:
                for host in task['hosts'].values():
                    logid = host.get('zuul_log_id')
                    if logid == 'skip':
                        skipped.add(task['task']['id'])
                        # The only tasks that should do this are:
                        assert "command should not stream" in host['stdout']

    # We should have skipped 2 tasks
    assert len(skipped) == 2
    for skip in skipped:
        # Verify that we did not attempt a connection
        assert f"Starting to log {skip}" not in console

    # We should have output due to copying from the result dict
    assert "This command should not stream 1" in output_txt
    assert "This command should not stream 2" in output_txt


if __name__ == "__main__":
    main()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.model import EventFilter, TriggerEvent


class ZuulEventFilter(EventFilter):
    def __init__(self, connection_name, trigger, types=[], pipelines=[],
                 debug=None):
        EventFilter.__init__(self, connection_name, trigger, debug)

        self._types = [x.pattern for x in types]
        self._pipelines = [x.pattern for x in pipelines]
        self.types = types
        self.pipelines = pipelines

    def __repr__(self):
        ret = '<ZuulEventFilter'
        ret += ' connection: %s' % self.connection_name

        if self._types:
            ret += ' types: %s' % ', '.join(self._types)
        if self._pipelines:
            ret += ' pipelines: %s' % ', '.join(self._pipelines)
        ret += '>'

        return ret

    def matches(self, event, change):
        if not super().matches(event, change):
            return False

        # event types are ORed
        matches_type = False
        for etype in self.types:
            if etype.match(event.type):
                matches_type = True
        if self.types and not matches_type:
            return False

        # pipelines are ORed
        matches_pipeline = False
        for epipe in self.pipelines:
            if epipe.match(event.pipeline_name):
                matches_pipeline = True
        if self.pipelines and not matches_pipeline:
            return False

        return True


class ZuulTriggerEvent(TriggerEvent):
    def __init__(self):
        super(ZuulTriggerEvent, self).__init__()
        self.pipeline_name = None
        self.image_names = None
        self.image_upload_uuid = None

    def toDict(self):
        d = super().toDict()
        d["pipeline_name"] = self.pipeline_name
        d["image_names"] = self.image_names
        d["image_upload_uuid"] = self.image_upload_uuid
        return d

    def updateFromDict(self, d):
        super().updateFromDict(d)
        self.pipeline_name = d["pipeline_name"]
        self.image_names = d.get("image_names")
        self.image_upload_uuid = d.get("image_upload_uuid")



# Copyright 2021 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE(ianw): 2021-08-30 we are testing whitespace things that trigger
# flake8, for now there is no way to turn of specific tests on a
# per-file basis.

# flake8: noqa

import textwrap

from zuul.lib.dependson import find_dependency_headers

from tests.base import BaseTestCase


class TestDependsOnParsing(BaseTestCase):

    def test_depends_on_parsing(self):
        msg = textwrap.dedent('''\
        This is a sample review subject

        Review text

        Depends-On:https://this.is.a.url/1
        Depends-On: https://this.is.a.url/2  
        Depends-On:    https://this.is.a.url/3
        Depends-On:	  https://this.is.a.url/4	
        ''')
        r = find_dependency_headers(msg)

        self.assertListEqual(r,
                             ['https://this.is.a.url/1',
                              'https://this.is.a.url/2',
                              'https://this.is.a.url/3',
                              'https://this.is.a.url/4'])



# Copyright 2020 BMW Group
# Copyright 2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import time
import zlib

from kazoo.exceptions import NoNodeError, NotEmptyError
from kazoo.retry import KazooRetry

from zuul.zk.locks import locked, SessionAwareLock
from zuul.zk.zkobject import LocalZKContext, ZKContext
from zuul.zk import sharding


class BlobStore:
    _retry_interval = 5
    lock_grace_period = 300
    data_root = "/zuul/cache/blob/data"
    lock_root = "/zuul/cache/blob/lock"

    def __init__(self, context):
        self.context = context

    def _getRootPath(self, key):
        return f"{self.data_root}/{key[0:2]}/{key}"

    def _getPath(self, key):
        root = self._getRootPath(key)
        return f"{root}/data"

    def _getFlagPath(self, key):
        root = self._getRootPath(key)
        return f"{root}/complete"

    def _getLockPath(self, key):
        return f"{self.lock_root}/{key}"

    def _retry(self, context, func, *args, max_tries=-1, **kw):
        kazoo_retry = KazooRetry(max_tries=max_tries,
                                 interrupt=context.shouldAbortRetry,
                                 delay=self._retry_interval, backoff=0,
                                 ignore_expire=False)
        try:
            return kazoo_retry(func, *args, **kw)
        except InterruptedError:
            pass

    @staticmethod
    def _retryableLoad(context, key, path, flag):
        if not context.client.exists(flag):
            raise KeyError(key)
        with sharding.BufferedShardReader(context.client, path) as stream:
            data = zlib.decompress(stream.read())
            context.cumulative_read_time += stream.cumulative_read_time
            context.cumulative_read_objects += 1
            context.cumulative_read_znodes += stream.znodes_read
            context.cumulative_read_bytes += stream.bytes_read
        return data, stream.bytes_read

    def get(self, key):
        path = self._getPath(key)
        flag = self._getFlagPath(key)

        if self.context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")

        data, compressed_size = self._retry(self.context, self._retryableLoad,
                                            self.context, key, path, flag)
        return data

    def _checkKey(self, key):
        # This returns whether the key is in the store.  If it is in
        # the store, it also touches the flag file so that the cleanup
        # routine can know the last time an entry was used.  Because
        # of the critical section between the get and delete calls in
        # the delete method, this must be called with the key lock.
        flag = self._getFlagPath(key)

        if self.context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")

        ret = self._retry(self.context, self.context.client.exists,
                          flag)
        if not ret:
            return False
        self._retry(self.context, self.context.client.set,
                    flag, b'')
        return True

    @staticmethod
    def _retryableSave(context, path, flag, data):
        with sharding.BufferedShardWriter(context.client, path) as stream:
            stream.truncate(0)
            stream.write(zlib.compress(data))
            stream.flush()
            context.client.ensure_path(flag)
            context.cumulative_write_time += stream.cumulative_write_time
            context.cumulative_write_objects += 1
            context.cumulative_write_znodes += stream.znodes_written
            context.cumulative_write_bytes += stream.bytes_written
        return stream.bytes_written

    def put(self, data):
        if isinstance(self.context, LocalZKContext):
            return None

        if self.context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")

        hasher = hashlib.sha256()
        hasher.update(data)
        key = hasher.hexdigest()

        path = self._getPath(key)
        flag = self._getFlagPath(key)
        lock_path = self._getLockPath(key)
        with locked(
                SessionAwareLock(
                    self.context.client,
                    lock_path),
                blocking=True
        ) as lock:
            if self._checkKey(key):
                return key

            # make a new context based on the old one
            with ZKContext(self.context.client, lock,
                           self.context.stop_event,
                           self.context.log) as locked_context:
                self._retry(
                    locked_context,
                    self._retryableSave,
                    locked_context, path, flag, data)
                self.context.updateStatsFromOtherContext(locked_context)
        return key

    def delete(self, key, ltime):
        path = self._getRootPath(key)
        flag = self._getFlagPath(key)
        lock_path = self._getLockPath(key)
        if self.context.sessionIsInvalid():
            raise Exception("ZooKeeper session or lock not valid")
        deleted = False
        try:
            with locked(
                    SessionAwareLock(
                        self.context.client,
                        lock_path),
                    blocking=True
            ) as lock:
                # make a new context based on the old one
                with ZKContext(self.context.client, lock,
                               self.context.stop_event,
                               self.context.log) as locked_context:

                    # Double check that it hasn't been used since we
                    # decided to delete it
                    data, zstat = self._retry(locked_context,
                                              self.context.client.get,
                                              flag)
                    if zstat.last_modified_transaction_id < ltime:
                        self._retry(locked_context, self.context.client.delete,
                                    path, recursive=True)
                        deleted = True
        except NoNodeError:
            raise KeyError(key)
        if deleted:
            try:
                lock_path = self._getLockPath(key)
                self._retry(self.context, self.context.client.delete,
                            lock_path)
            except Exception:
                self.context.log.exception(
                    "Error deleting lock path %s:", lock_path)

    def __iter__(self):
        try:
            hashdirs = self.context.client.get_children(self.data_root)
        except NoNodeError:
            return

        for hashdir in hashdirs:
            try:
                for key in self.context.client.get_children(
                        f'{self.data_root}/{hashdir}'):
                    yield key
            except NoNodeError:
                pass

    def __len__(self):
        return len([x for x in self])

    def getKeysLastUsedBefore(self, ltime):
        ret = set()
        for key in self:
            flag = self._getFlagPath(key)
            data, zstat = self._retry(self.context, self.context.client.get,
                                      flag)
            if zstat.last_modified_transaction_id < ltime:
                ret.add(key)
        return ret

    def cleanupLockDirs(self, start_ltime, live_blobs):
        # This cleanup was not present in an earlier version of Zuul,
        # therefore lock directory entries could grow without bound.
        # If there are too many entries, we won't be able to list them
        # in order to delete them and the connection will be closed.
        # Before we proceed, make sure that isn't the case here.
        # The size of the packet will be:
        # (num_children * (hash_length + int_size))
        # (num_children * (64 + 4)) = num_children * 68
        max_children = sharding.NODE_BYTE_SIZE_LIMIT / 68
        zstat = self._retry(self.context, self.context.client.exists,
                            self.lock_root)
        if not zstat:
            # Lock root does not exist
            return
        if zstat.children_count > max_children:
            self.context.log.error(
                "Unable to cleanup blob store lock directory "
                "due to too many lock znodes.")
            return

        # We're not retrying here just in case we calculate wrong and
        # we get a ZK disconnection.  We could change this in the
        # future when we're no longer worried about the number of
        # children.
        keys = self.context.client.get_children(self.lock_root)
        for key in keys:
            if key in live_blobs:
                # No need to check a live blob
                continue
            path = self._getLockPath(key)
            zstat = self._retry(self.context, self.context.client.exists,
                                path)
            if not zstat:
                continue
            # Any lock dir that is not for a live blob is either:
            # 1) leaked
            #    created time will be old, okay to delete
            # 2) is for a newly created blob
            #    created time will be new, not okay to delete
            # 3) is for a previously used blob about to be re-created
            #    created time will be old, not okay to delete
            # We can not detect case 3, but it is unlikely to happen,
            # and if it does, the locked context manager in the put
            # method will recreate the lock directory after we delete
            # it.
            now = time.time()
            if ((zstat.created > now - self.lock_grace_period) or
                (zstat.creation_transaction_id > start_ltime)):
                # The lock was recently created, we may have caught it
                # while it's being created.
                continue
            try:
                self.context.log.debug("Deleting unused key dir: %s", path)
                self.context.client.delete(path)
            except NotEmptyError:
                # It may still be in use
                pass
            except NoNodeError:
                pass
            except Exception:
                self.context.log.exception(
                    "Error deleting lock path %s:", path)

    def cleanup(self, start_ltime, live_blobs):
        # get the set of blob keys unused since the start time
        # (ie, we have already filtered any newly added keys)
        unused_blobs = self.getKeysLastUsedBefore(start_ltime)
        # remove the current refences
        unused_blobs -= live_blobs
        # delete what's left
        for key in unused_blobs:
            self.context.log.debug("Deleting unused blob: %s", key)
            self.delete(key, start_ltime)
        self.cleanupLockDirs(start_ltime, live_blobs)



# Copyright (C) 2018 Red Hat
# Copyright 2023, 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


class FakeCoreClient:
    def __init__(self):
        self.namespaces = {}
        self._pod_requests = []

        ns_body = {
            'apiVersion': 'v1',
            'kind': 'Namespace',
            'metadata': {
                'name': 'default',
            }
        }
        self.create_namespace(ns_body)

        class FakeApi:
            class configuration:
                host = "http://localhost:8080"
                verify_ssl = False
        self.api_client = FakeApi()

    def list_namespace(self):
        class FakeNamespaces:
            items = list(self.namespaces.values())
        return FakeNamespaces

    def create_namespace(self, ns_body):
        class FakeNamespace:
            class metadata:
                name = ns_body['metadata']['name']
                labels = ns_body['metadata'].get('labels')
            _secrets = {}
            _pods = {}
        self.namespaces[FakeNamespace.metadata.name] = FakeNamespace
        return FakeNamespace

    def delete_namespace(self, name, body):
        if name not in self.namespaces:
            raise RuntimeError("Unknown namespace %s" % name)
        del self.namespaces[name]

    def create_namespaced_service_account(self, ns, sa_body):
        return

    def read_namespaced_service_account(self, user, ns):
        class FakeSA:
            class secret:
                name = "fake"
        FakeSA.secrets = [FakeSA.secret]
        return FakeSA

    def create_namespaced_secret(self, ns, secret_body):
        namespace = self.namespaces[ns]
        name = secret_body['metadata']['name']

        class FakeSecret:
            class metadata:
                name = secret_body['metadata']['name']
            data = secret_body.get('data')
        if secret_body['type'] == 'kubernetes.io/service-account-token':
            FakeSecret.data = {
                'ca.crt': 'ZmFrZS1jYQ==', 'token': 'ZmFrZS10b2tlbg=='
            }
        namespace._secrets[name] = FakeSecret

    def read_namespaced_secret(self, name, ns):
        namespace = self.namespaces[ns]
        return namespace._secrets[name]

    def create_namespaced_pod(self, ns, pod_body):
        namespace = self.namespaces[ns]
        name = pod_body['metadata']['name']

        class FakePod:
            class status:
                phase = "Running"

            class metadata:
                name = pod_body['metadata']['name']
                labels = pod_body['metadata'].get('labels')
            spec = pod_body.get('spec')
        namespace._pods[name] = FakePod

    def read_namespaced_pod(self, name, ns):
        namespace = self.namespaces[ns]
        return namespace._pods[name]

    def list_namespaced_pod(self, ns):
        namespace = self.namespaces[ns]

        class PodList:
            items = namespace._pods.values()
        return PodList


class FakeRbacClient:
    def create_namespaced_role(self, ns, role_body):
        return

    def create_namespaced_role_binding(self, ns, role_binding_body):
        return


class FakeDynamicClient:
    def __init__(self, core_client, openshift):
        class FakeResources:
            def get(self, api_version, kind):
                if kind == 'ProjectRequest':
                    if not openshift:
                        raise Exception("Not an OpenShift server")

                    class FakeProjects:
                        def create(self, body):
                            new_body = copy.deepcopy(body)
                            new_body['apiVersion'] = 'v1'
                            new_body['kind'] = 'Namespace'
                            core_client.create_namespace(new_body)
                    return FakeProjects()
                if kind == 'Project':
                    if not openshift:
                        raise Exception("Not an OpenShift server")

                    class FakeProject:
                        def get(self):

                            class FakeProjects:
                                items = list(core_client.namespaces.values())
                            return FakeProjects

                        def delete(self, name):
                            core_client.delete_namespace(name, None)
                    return FakeProject()

        self.resources = FakeResources()



# Copyright 2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid

import botocore
import boto3

from zuul.driver.aws.awsendpoint import AwsProviderEndpoint


def make_import_snapshot_stage_1(task_id, user_bucket, tags):
    return {
        'Architecture': 'x86_64',
        'ImportTaskId': f'import-snap-{task_id}',
        'Progress': '19',
        'SnapshotTaskDetail': {'DiskImageSize': 355024384.0,
                               'Format': 'VMDK',
                               'Status': 'active',
                               'UserBucket': user_bucket},
        'Status': 'active',
        'StatusMessage': 'converting',
        'Tags': tags,
    }


def make_import_snapshot_stage_2(task_id, snap_id, task):
    # Make a unique snapshot id that's different than the task id.
    return {
        'ImportTaskId': f'import-snap-{task_id}',
        'SnapshotTaskDetail': {'DiskImageSize': 355024384.0,
                               'Format': 'VMDK',
                               'SnapshotId': snap_id,
                               'Status': 'completed',
                               'UserBucket':
                               task['SnapshotTaskDetail']['UserBucket']},
        'Status': 'completed',
        'Tags': task['Tags'],
    }


def make_import_image_stage_1(task_id, user_bucket, tags):
    return {
        'Architecture': 'x86_64',
        'ImportTaskId': f'import-ami-{task_id}',
        'Progress': '19',
        'SnapshotDetails': [{'DiskImageSize': 355024384.0,
                             'Format': 'VMDK',
                             'Status': 'active',
                             'UserBucket': user_bucket}],
        'Status': 'active',
        'StatusMessage': 'converting',
        'Tags': tags,
    }


def make_import_image_stage_2(task_id, image_id, snap_id, task):
    # Make a unique snapshot id that's different than the task id.
    return {
        'Architecture': 'x86_64',
        'BootMode': 'legacy_bios',
        'ImageId': image_id,
        'ImportTaskId': f'import-ami-{task_id}',
        'LicenseType': 'BYOL',
        'Platform': 'Linux',
        'SnapshotDetails': [{'DeviceName': '/dev/sda1',
                             'DiskImageSize': 355024384.0,
                             'Format': 'VMDK',
                             'SnapshotId': snap_id,
                             'Status': 'completed',
                             'UserBucket':
                             task['SnapshotDetails'][0]['UserBucket']}],
        'Status': 'completed',
        'Tags': task['Tags'],
    }


class ImportSnapshotTaskPaginator:
    log = logging.getLogger("nodepool.FakeAws")

    def __init__(self, fake):
        self.fake = fake

    def paginate(self, **kw):
        tasks = list(self.fake.tasks.values())
        tasks = [t for t in tasks if 'import-snap' in t['ImportTaskId']]
        if 'ImportTaskIds' in kw:
            tasks = [t for t in tasks
                     if t['ImportTaskId'] in kw['ImportTaskIds']]
        # A page of tasks
        ret = [{'ImportSnapshotTasks': tasks}]

        # Move the task along
        for task in tasks:
            if task['Status'] != 'completed':
                self.fake.finish_import_snapshot(task)
        return ret


class ImportImageTaskPaginator:
    log = logging.getLogger("nodepool.FakeAws")

    def __init__(self, fake):
        self.fake = fake

    def paginate(self, **kw):
        tasks = list(self.fake.tasks.values())
        tasks = [t for t in tasks if 'import-ami' in t['ImportTaskId']]
        if 'ImportTaskIds' in kw:
            tasks = [t for t in tasks
                     if t['ImportTaskId'] in kw['ImportTaskIds']]
            if not tasks:
                raise botocore.exceptions.ClientError(
                    {'Error': {'Code': 'InvalidConversionTaskId.Malformed'}},
                    'DescribeImportImageTasks')
        # A page of tasks
        ret = [{'ImportImageTasks': tasks}]

        # Move the task along
        for task in tasks:
            if task['Status'] != 'completed':
                self.fake.finish_import_image(task)
        return ret


class FakeAws:
    log = logging.getLogger("nodepool.FakeAws")

    def __init__(self):
        self.tasks = {}
        self.ec2 = boto3.resource('ec2', region_name='us-east-1')
        self.ec2_client = boto3.client('ec2', region_name='us-east-1')
        self.fail_import_count = 0

    def import_snapshot(self, *args, **kw):
        while self.fail_import_count:
            self.fail_import_count -= 1
            raise botocore.exceptions.ClientError(
                {'Error': {'Code': 'ResourceCountLimitExceeded'}},
                'ImportSnapshot')
        task_id = uuid.uuid4().hex
        task = make_import_snapshot_stage_1(
            task_id,
            kw['DiskContainer']['UserBucket'],
            kw['TagSpecifications'][0]['Tags'])
        self.tasks[task_id] = task
        return task

    def finish_import_snapshot(self, task):
        task_id = task['ImportTaskId'].split('-')[-1]

        # Make a Volume to simulate the import finishing
        volume = self.ec2_client.create_volume(
            Size=80,
            AvailabilityZone='us-east-1')
        snap_id = self.ec2_client.create_snapshot(
            VolumeId=volume['VolumeId'],
        )["SnapshotId"]

        t2 = make_import_snapshot_stage_2(task_id, snap_id, task)
        self.tasks[task_id] = t2
        return snap_id

    def import_image(self, *args, **kw):
        while self.fail_import_count:
            self.fail_import_count -= 1
            raise botocore.exceptions.ClientError(
                {'Error': {'Code': 'ResourceCountLimitExceeded'}},
                'ImportImage')
        task_id = uuid.uuid4().hex
        task = make_import_image_stage_1(
            task_id,
            kw['DiskContainers'][0]['UserBucket'],
            kw['TagSpecifications'][0]['Tags'])
        self.tasks[task_id] = task
        return task

    def finish_import_image(self, task):
        task_id = task['ImportTaskId'].split('-')[-1]

        # Make an AMI to simulate the import finishing
        reservation = self.ec2_client.run_instances(
            ImageId="ami-12c6146b", MinCount=1, MaxCount=1)
        instance = reservation["Instances"][0]
        instance_id = instance["InstanceId"]

        response = self.ec2_client.create_image(
            InstanceId=instance_id,
            Name=f'import-ami-{task_id}',
        )

        image_id = response["ImageId"]
        self.ec2_client.describe_images(ImageIds=[image_id])["Images"][0]

        volume = self.ec2_client.create_volume(
            Size=80,
            AvailabilityZone='us-east-1')
        snap_id = self.ec2_client.create_snapshot(
            VolumeId=volume['VolumeId'],
            Description=f'imported volume import-ami-{task_id}',
        )["SnapshotId"]

        t2 = make_import_image_stage_2(task_id, image_id, snap_id, task)
        self.tasks[task_id] = t2
        return (image_id, snap_id)

    def change_snapshot_id(self, task, snapshot_id):
        # Given a task, update its snapshot id; the moto
        # register_image mock doesn't honor the snapshot_id we pass
        # in.
        task_id = task['ImportTaskId'].split('-')[-1]
        self.tasks[task_id]['SnapshotTaskDetail']['SnapshotId'] = snapshot_id

    def get_paginator(self, name):
        if name == 'describe_import_image_tasks':
            return ImportImageTaskPaginator(self)
        if name == 'describe_import_snapshot_tasks':
            return ImportSnapshotTaskPaginator(self)
        raise NotImplementedError()

    def _listAmis(self):
        return list(self.ec2.images.filter())

    def _listSnapshots(self):
        return list(self.ec2.snapshots.filter())


class FakeAwsProviderEndpoint(AwsProviderEndpoint):
    IMAGE_UPLOAD_SLEEP = 1

    def _getClients(self):
        super()._getClients()

        # Note: boto3 doesn't handle ipv6 addresses correctly
        # when in fake mode so we need to intercept the
        # run_instances call and validate the args we supply.
        def _fake_run_instances(*args, **kwargs):
            self.__testcase.run_instances_calls.append(kwargs)
            if self.__testcase.run_instances_exception:
                raise self.__testcase.run_instances_exception
            return self.ec2_client.run_instances_orig(*args, **kwargs)

        # Moto doesn't handle all features correctly (e.g.
        # instance-requirements, volume attributes) when creating
        # fleet in fake mode, we need to intercept the create_fleet
        # call and validate the args we supply. Results are also
        # intercepted for validate instance attributes
        def _fake_create_fleet(*args, **kwargs):
            self.__testcase.create_fleet_calls.append(kwargs)
            if self.__testcase.create_fleet_exception:
                raise self.__testcase.create_fleet_exception
            result = self.ec2_client.create_fleet_orig(*args, **kwargs)
            self.__testcase.create_fleet_results.append(result)
            return result

        def _fake_allocate_hosts(*args, **kwargs):
            if self.__testcase.allocate_hosts_exception:
                raise self.__testcase.allocate_hosts_exception
            return self.ec2_client.allocate_hosts_orig(*args, **kwargs)

        # The ImdsSupport parameter isn't handled by moto
        def _fake_register_image(*args, **kwargs):
            self.__testcase.register_image_calls.append(kwargs)
            return self.ec2_client.register_image_orig(*args, **kwargs)

        # Just so we can count how many times it's called
        def _fake_copy_image(*args, **kwargs):
            self.__testcase.copy_image_calls.append(kwargs)
            return self.ec2_client.copy_image_orig(*args, **kwargs)

        def _fake_get_paginator(*args, **kwargs):
            try:
                return self.__testcase.fake_aws.get_paginator(*args, **kwargs)
            except NotImplementedError:
                return self.ec2_client.get_paginator_orig(*args, **kwargs)

        self.ec2_client.run_instances_orig = self.ec2_client.run_instances
        self.ec2_client.run_instances = _fake_run_instances
        self.ec2_client.create_fleet_orig = self.ec2_client.create_fleet
        self.ec2_client.create_fleet = _fake_create_fleet
        self.ec2_client.allocate_hosts_orig = self.ec2_client.allocate_hosts
        self.ec2_client.allocate_hosts = _fake_allocate_hosts
        self.ec2_client.register_image_orig = self.ec2_client.register_image
        self.ec2_client.register_image = _fake_register_image
        self.ec2_client.copy_image_orig = self.ec2_client.copy_image
        self.ec2_client.copy_image = _fake_copy_image
        self.ec2_client.import_snapshot = \
            self.__testcase.fake_aws.import_snapshot
        self.ec2_client.import_image = \
            self.__testcase.fake_aws.import_image
        self.ec2_client.get_paginator_orig = self.ec2_client.get_paginator
        self.ec2_client.get_paginator = _fake_get_paginator

        # moto does not mock service-quotas, so we do it ourselves:
        def _fake_get_service_quota(ServiceCode, QuotaCode, *args, **kwargs):
            if ServiceCode == 'ec2':
                qdict = self.__ec2_quotas
            elif ServiceCode == 'ebs':
                qdict = self.__ebs_quotas
            else:
                raise NotImplementedError(
                    f"Quota code {ServiceCode} not implemented")
            return {'Quota': {'Value': qdict.get(QuotaCode)}}
        self.aws_quotas.get_service_quota = _fake_get_service_quota

        def _fake_list_service_quotas(ServiceCode, *args, **kwargs):
            if ServiceCode == 'ec2':
                qdict = self.__ec2_quotas
            elif ServiceCode == 'ebs':
                qdict = self.__ebs_quotas
            else:
                raise NotImplementedError(
                    f"Quota code {ServiceCode} not implemented")
            quotas = []
            for code, value in qdict.items():
                quotas.append(
                    {'Value': value, 'QuotaCode': code}
                )
            return {'Quotas': quotas}
        self.aws_quotas.list_service_quotas = _fake_list_service_quotas



# Copyright 2024 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import json
import logging
import threading

import mmh3
from kazoo.exceptions import NoNodeError

from zuul.model import (
    NodesetRequest,
    ProviderNode,
    ProviderNodeAssignment,
    ProviderNodeSnapshot,
    QuotaInformation,
)
from zuul.zk.cache import ZuulTreeCache
from zuul.zk.zkobject import ZKContext


def _dictToBytes(data):
    return json.dumps(data).encode("utf-8")


def _bytesToDict(raw_data):
    return json.loads(raw_data.decode("utf-8"))


def launcher_score(name, item):
    return mmh3.hash(f"{name}-{item.uuid}", signed=False)


class LockableZKObjectCache(ZuulTreeCache):

    def __init__(self, zk_client, updated_event, root, items_path,
                 locks_path, zkobject_class):
        self.updated_event = updated_event
        self.items_path = items_path
        self.locks_path = locks_path
        self.zkobject_class = zkobject_class
        super().__init__(zk_client, root)

    def _parsePath(self, path):
        if not path.startswith(self.root):
            return None
        path = path[len(self.root) + 1:]
        parts = path.split('/')
        # We are interested in requests with a parts that look like:
        # ([<self.items_path>, <self.locks_path>], <uuid>, ...)
        if len(parts) < 2:
            return None
        return parts

    def parsePath(self, path):
        key = None
        fetch = False
        parts = self._parsePath(path)
        if parts is None:
            return (key, fetch)
        if len(parts) < 2:
            return (key, fetch)

        object_type, item_uuid, *_ = parts
        if object_type == self.items_path:
            key = (item_uuid,)
            fetch = True
        elif object_type == self.locks_path:
            key = None
            if len(parts) == 3:
                fetch = True

        return (key, fetch)

    def preCacheHook(self, event, exists, data=None, stat=None):
        parts = self._parsePath(event.path)
        if parts is None:
            return

        # Expecting (<self.locks_path>, <uuid>, <lock>,)
        if len(parts) != 3:
            return

        object_type, request_uuid, contender, *_ = parts
        if object_type != self.locks_path:
            return

        key = (request_uuid,)
        request = self._cached_objects.get(key)

        if not request:
            return

        if exists:
            request._lock_contenders[contender] = data
        else:
            request._lock_contenders.pop(contender, None)

        is_locked = bool(request._lock_contenders)
        lock_holder = None
        if is_locked:
            # This is a very simplified contender lock algorithm that
            # only works for exclusive locks.
            sorted_names = sorted(request._lock_contenders.keys(),
                                  key=lambda x: x[-10:])
            if sorted_names:
                lock_holder = request._lock_contenders[sorted_names[0]]
                if lock_holder:
                    lock_holder = lock_holder.decode('utf8')
                else:
                    lock_holder = None
        update_args = {}
        if request.is_locked != is_locked:
            update_args['is_locked'] = is_locked
        if request.lock_holder != lock_holder:
            update_args['lock_holder'] = lock_holder
        if update_args:
            request._set(**update_args)
            if self.updated_event:
                self.updated_event()

    def postCacheHook(self, event, data, stat, key, obj):
        if self.updated_event:
            self.updated_event()

    def objectFromRaw(self, key, data, zstat):
        return self.zkobject_class._fromRaw(
            self._zk_context, data, zstat, None)

    def updateFromRaw(self, obj, key, data, zstat):
        obj._updateFromRaw(self._zk_context, data, zstat, None)

    def getItem(self, item_id):
        self.ensureReady()
        return self._cached_objects.get((item_id,))

    def getItems(self):
        # get a copy of the values view to avoid runtime errors in the event
        # the _cached_nodes dict gets updated while iterating
        self.ensureReady()
        return list(self._cached_objects.values())

    def getKeys(self):
        self.ensureReady()
        return set(self._cached_objects.keys())


class RequestCache(LockableZKObjectCache):

    def preCacheHook(self, event, exists, data=None, stat=None):
        parts = self._parsePath(event.path)
        if parts is None:
            return

        # Expecting (<self.locks_path>, <uuid>, <lock>,)
        # or (<self.items_path>, <uuid>, revision,)
        if len(parts) != 3:
            return

        object_type, request_uuid, *_ = parts
        key = (request_uuid,)
        request = self._cached_objects.get(key)

        if not request:
            return

        if object_type == self.locks_path:
            request._set(is_locked=exists)
        elif data is not None:
            request._revision._updateFromRaw(
                self._zk_context, data, stat, None)
            return self.STOP_OBJECT_UPDATE


class NodeCache(LockableZKObjectCache):
    def __init__(self, *args, **kw):
        # The states we're interested in
        self._quota_states_used = ProviderNode.ALLOCATED_STATES
        self._quota_states_requested = (
            ProviderNode.ALLOCATED_STATES + (ProviderNode.State.REQUESTED,)
        )
        # Key -> quota, for each cached object
        self._cached_quota_used = {}
        self._cached_quota_requested = {}
        # Provider canonical name -> quota, for each provider
        self._provider_quota_used = collections.defaultdict(
            lambda: QuotaInformation())
        self._provider_quota_requested = collections.defaultdict(
            lambda: QuotaInformation())
        super().__init__(*args, **kw)

    def _handleQuota(self, quota_states, quota_cache, provider_cache,
                     event, data, stat, key, obj):
        # Have we previously added quota for this object?
        old_quota = quota_cache.get(key)
        if key in self._cached_objects:
            new_quota = obj.quota
        else:
            new_quota = None
        # Now that we've established whether we should count the quota
        # based on object presence, take node state into account.
        if obj is None or obj.state not in quota_states:
            new_quota = None
        if new_quota != old_quota:
            # Add the new value first so if another thread races these
            # two operations, it sees us go over quota and not under.
            if new_quota is not None:
                provider_cache[obj.provider].add(new_quota)
            if old_quota is not None:
                provider_cache[obj.provider].subtract(old_quota)
            if new_quota is None:
                quota_cache.pop(key, None)
            else:
                quota_cache[key] = new_quota

    def preCacheHook(self, event, exists, data=None, stat=None):
        parts = self._parsePath(event.path)
        if parts is None:
            return

        # (<self.locks_path>, <uuid>, <lock>,)
        # (<self.items_path>, <uuid>,)
        # (<self.items_path>, <uuid>, snapshot,)
        # (<self.items_path>, <uuid>, snapshot-lock,)
        # (<self.items_path>, <uuid>, assignment,)
        if len(parts) >= 3:
            # Ignore anything related to snapshots
            if (parts[0] == ProviderNode.NODES_PATH and
                parts[2] in (ProviderNodeSnapshot.SNAPSHOT_PATH,
                             ProviderNodeSnapshot.SNAPSHOT_LOCK_PATH)):
                return self.STOP_OBJECT_UPDATE
            if (parts[0] == ProviderNode.NODES_PATH and
                parts[2] == ProviderNodeAssignment.ASSIGNMENT_PATH):
                key = (parts[1],)
                node = self._cached_objects.get(key)
                if not node:
                    return
                if exists:
                    node.assignment._updateFromRaw(
                        self._zk_context, data, stat, None)
                else:
                    node.assignment._clear()
                return self.STOP_OBJECT_UPDATE
        return super().preCacheHook(event, exists, data, stat)

    def postCacheHook(self, event, data, stat, key, obj):
        self._handleQuota(
            self._quota_states_used,
            self._cached_quota_used,
            self._provider_quota_used,
            event, data, stat, key, obj)
        self._handleQuota(
            self._quota_states_requested,
            self._cached_quota_requested,
            self._provider_quota_requested,
            event, data, stat, key, obj)
        super().postCacheHook(event, data, stat, key, obj)

    def getQuota(self, provider, include_requested=False):
        if include_requested:
            return self._provider_quota_requested[provider.canonical_name]
        else:
            return self._provider_quota_used[provider.canonical_name]


class LauncherApi:
    log = logging.getLogger("zuul.LauncherApi")

    def __init__(self, zk_client, component_registry, component_info,
                 event_callback, connection_filter=None):
        self.zk_client = zk_client
        self.component_registry = component_registry
        self.component_info = component_info
        self.event_callback = event_callback
        self.requests_cache = RequestCache(
            self.zk_client,
            self.event_callback,
            root=NodesetRequest.ROOT,
            items_path=NodesetRequest.REQUESTS_PATH,
            locks_path=NodesetRequest.LOCKS_PATH,
            zkobject_class=NodesetRequest)
        self.nodes_cache = NodeCache(
            self.zk_client,
            self.event_callback,
            root=ProviderNode.ROOT,
            items_path=ProviderNode.NODES_PATH,
            locks_path=ProviderNode.LOCKS_PATH,
            zkobject_class=ProviderNode)
        self.connection_filter = connection_filter
        self.stop_event = threading.Event()

    def stop(self):
        self.stop_event.set()
        self.requests_cache.stop()
        self.nodes_cache.stop()

    def getMatchingRequests(self):
        candidate_launchers = {
            c.hostname: c for c in self.component_registry.all("launcher")}
        candidate_names = set(candidate_launchers.keys())

        for request in sorted(self.requests_cache.getItems()):
            if request.hasLock():
                # We are holding a lock, so short-circuit here.
                yield request
            if request.is_locked:
                # Request is locked by someone else
                continue
            if request.state == NodesetRequest.State.TEST_HOLD:
                continue

            score_launchers = (
                set(request._lscores.keys()) if request._lscores else set())
            missing_scores = candidate_names - score_launchers
            if missing_scores or request._lscores is None:
                # (Re-)compute launcher scores
                request._set(_lscores={launcher_score(n, request): n
                                       for n in candidate_names})

            launcher_scores = sorted(request._lscores.items())
            # self.log.debug("Launcher scores: %s", launcher_scores)
            for score, launcher_name in launcher_scores:
                launcher = candidate_launchers.get(launcher_name)
                if not launcher:
                    continue
                if launcher.state != launcher.RUNNING:
                    continue
                if launcher.hostname == self.component_info.hostname:
                    yield request
                break

    def getNodesetRequest(self, request_id):
        return self.requests_cache.getItem(request_id)

    def getNodesetRequests(self):
        return sorted(self.requests_cache.getItems())

    def getMatchingProviderNodes(self):
        all_launchers = {
            c.hostname: c for c in self.component_registry.all("launcher")}

        for node in sorted(self.nodes_cache.getItems()):
            if node.hasLock():
                # We are holding a lock, so short-circuit here.
                yield node
            if node.is_locked:
                # Node is locked by someone else
                if node.state != node.State.SNAPSHOT:
                    # But in snapshot state, we may have work to do
                    continue

            candidate_launchers = {
                n: c for n, c in all_launchers.items()
                if not c.connection_filter
                or node.connection_name in c.connection_filter}
            candidate_names = set(candidate_launchers)
            if node._lscores is None:
                missing_scores = candidate_names
            else:
                score_launchers = set(node._lscores.keys())
                missing_scores = candidate_names - score_launchers

            if missing_scores or node._lscores is None:
                # (Re-)compute launcher scores
                node._set(_lscores={launcher_score(n, node): n
                                    for n in candidate_names})

            launcher_scores = sorted(node._lscores.items())

            for score, launcher_name in launcher_scores:
                launcher = candidate_launchers.get(launcher_name)
                if not launcher:
                    # Launcher is no longer online
                    continue
                if launcher.state != launcher.RUNNING:
                    continue
                if launcher.hostname == self.component_info.hostname:
                    yield node
                break

    def getProviderNode(self, node_id):
        return self.nodes_cache.getItem(node_id)

    def getProviderNodes(self):
        return self.nodes_cache.getItems()

    def createZKContext(self, lock=None, log=None):
        return ZKContext(
            self.zk_client, lock, self.stop_event, log or self.log)

    def cleanupNodes(self):
        # TODO: This method currently just performs some basic cleanup and
        # might need to be extended in the future.
        for node in self.getProviderNodes():
            if node.state != ProviderNode.State.USED:
                continue
                # FIXME: check if the node request still exists
            if node.is_locked:
                continue
            if self.getNodesetRequest(node.request_id):
                continue
            with self.createZKContext(None) as outer_ctx:
                if lock := node.acquireLock(outer_ctx):
                    with self.createZKContext(lock) as ctx:
                        try:
                            node.delete(ctx)
                        except NoNodeError:
                            # Node is already deleted
                            pass
                        finally:
                            node.releaseLock(ctx)



# Copyright (c) 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tests.base import ZuulTestCase


class TestPushRequirements(ZuulTestCase):
    config_file = 'zuul-push-reqs.conf'
    tenant_config_file = 'config/push-reqs/main.yaml'

    def test_push_requirements(self):
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_github.openFakePullRequest('org/project1', 'master', 'A')
        new_sha = A.head_sha
        A.setMerged("merging A")
        pevent = self.fake_github.getPushEvent(project='org/project1',
                                               ref='refs/heads/master',
                                               new_rev=new_sha)

        self.fake_github.emitEvent(pevent)

        self.waitUntilSettled()

        # All but one pipeline should be skipped
        self.assertEqual(1, len(self.builds))
        self.assertEqual('pushhub', self.builds[0].pipeline)
        self.assertEqual('org/project1', self.builds[0].project)

        # Make a gerrit change, and emit a ref-updated event
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        self.fake_gerrit.addEvent(B.getRefUpdatedEvent())
        B.setMerged()
        self.waitUntilSettled()

        # All but one pipeline should be skipped, increasing builds by 1
        self.assertEqual(2, len(self.builds))
        self.assertEqual('pushgerrit', self.builds[1].pipeline)
        self.assertEqual('org/project2', self.builds[1].project)

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()



# Copyright 2024 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import signal

import zuul.cmd
from zuul.launcher.server import COMMANDS, Launcher


class LauncherApp(zuul.cmd.ZuulDaemonApp):
    app_name = 'launcher'
    app_description = 'The Zuul launcher.'
    launcher = None

    def __init__(self):
        super().__init__()

    def createParser(self):
        parser = super().createParser()
        self.addSubCommands(parser, COMMANDS)
        return parser

    def exit_handler(self, signum, frame):
        if self.launcher:
            self.launcher.stop()
            self.launcher.join()

    def run(self):
        self.handleCommands()

        self.setup_logging('launcher', 'log_config')
        self.log = logging.getLogger('zuul.launcher')

        self.configure_connections(providers=True)
        self.launcher = Launcher(self.config, self.connections)
        self.launcher.start()

        if self.args.nodaemon:
            signal.signal(signal.SIGTERM, self.exit_handler)

        while True:
            try:
                self.launcher.join()
                break
            except KeyboardInterrupt:
                print("Ctrl + C: asking launcher to exit nicely...\n")
                self.exit_handler(signal.SIGINT, None)


def main():
    LauncherApp().main()



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import os.path

from zuul.executor.sensors import SensorInterface


class ProcessSensor(SensorInterface):
    log = logging.getLogger("zuul.executor.sensor.process")

    def __init__(self, statsd, base_key, config=None):
        super().__init__(statsd, base_key)
        # The executor and ansible require a number of processes to function
        # minimally: the executor itself, ansible, ssh control persistence,
        # ssh and so on. Set a minimum of room for 10 processes before we
        # stop.
        self._min_headroom = 10
        self._safety_factor = 0.05
        self._uid = os.getuid()
        self._pid_max = self._get_pid_max()
        self._root_cgroup_max_file = '/sys/fs/cgroup/pids.max'
        self._root_cgroup_cur_file = '/sys/fs/cgroup/pids.current'
        # This appears to be systemd specific behavior with cgroups that
        # reflects the ulimit values. This way we don't need to have a
        # separate system for ulimit checking.
        self._user_cgroup_max_file = f'/sys/fs/cgroup/user.slice' \
                                     f'/user-{self._uid}.slice/pids.max'
        self._user_cgroup_cur_file = f'/sys/fs/cgroup/user.slice' \
                                     f'/user-{self._uid}.slice/pids.current'

    def _get_pid_max(self):
        # Default for x86_64
        default = 2 ** 22
        path = '/proc/sys/kernel/pid_max'
        if os.path.exists(path):
            with open(path) as f:
                s = f.read().strip()
                try:
                    i = int(s)
                except ValueError:
                    self.log.exception('Unable to determine pid_max')
                    i = default
                return i
        else:
            return default

    def isOk(self):
        # Processes running in the root cgroup won't have these values
        # but containers do.
        # If no max is found assume pid_max. If no current usage is found
        # assume 1 for the current process.
        root_max = self._get_root_cgroup_max() or self._pid_max
        root_current = self._get_root_cgroup_current() or 1
        # Processes running under systemd will have these values.
        user_max = self._get_user_slice_max() or self._pid_max
        user_current = self._get_user_slice_current() or 1

        limit = min(root_max, user_max)
        usage = max(root_current, user_current)
        min_headroom = limit * self._safety_factor
        if min_headroom < self._min_headroom:
            min_headroom = self._min_headroom
        # This shouldn't ever be negative but I'm not sure if you can reduce
        # cgroup limits below the current usage at runtime.
        headroom = max(limit - usage, 0)

        if self.statsd:
            self.statsd.gauge(self.base_key + '.max_process',
                              limit)
            self.statsd.gauge(self.base_key + '.cur_process',
                              usage)

        if min_headroom >= headroom:
            return False, f'high process utilization: {usage} max: {limit}'
        return True, f'process utilization: {usage} max: {limit}'

    def _get_root_cgroup_max(self):
        return self._get_cgroup_value(self._root_cgroup_max_file)

    def _get_root_cgroup_current(self):
        return self._get_cgroup_value(self._root_cgroup_cur_file)

    def _get_user_slice_max(self):
        return self._get_cgroup_value(self._user_cgroup_max_file)

    def _get_user_slice_current(self):
        return self._get_cgroup_value(self._user_cgroup_cur_file)

    def _get_cgroup_value(self, path):
        if os.path.exists(path):
            with open(path) as f:
                s = f.read().strip()
                if s == 'max':
                    return self._pid_max
                else:
                    try:
                        i = int(s)
                    except ValueError:
                        self.log.exception('Unable to convert cgroup '
                                           'process value')
                        i = None
                    return i
        else:
            return None



# Copyright 2025 Acme Gating, LLC
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import importlib.resources

from ansible.plugins.action import ActionBase
from ansible.utils.vars import merge_hash

START_SCRIPT = """
(Add-Type -Path "{remote_cs_path}")
[WinZuulConsole]::Main($args)
"""


class ActionModule(ActionBase):
    TRANSFERS_FILES = True

    def run(self, tmp=None, task_vars=None):
        results = super(ActionModule, self).run(tmp, task_vars)

        # Copy our csharp server code over
        local_path = str(importlib.resources.files('zuul').joinpath(
            'ansible/win_zuul_console.cs'))
        remote_cs_path = self._connection._shell.join_path(
            self._connection._shell.tmpdir,
            'win_zuul_console.cs')
        self._transfer_file(local_path, remote_cs_path)

        # Copy the bootstrap script
        remote_ps_path = self._connection._shell.join_path(
            self._connection._shell.tmpdir,
            'win_zuul_console_start.ps1')
        self._transfer_data(
            remote_ps_path,
            START_SCRIPT.format(remote_cs_path=remote_cs_path))
        self._task.args['_zuul_console_exec_path'] = remote_ps_path

        # Run the module
        results = merge_hash(results, self._execute_module(
            module_name='win_zuul_console', task_vars=task_vars))
        return results



# Copyright 2020 BMW Group
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
from contextlib import suppress
import time
import zlib

from zuul.zk.components import COMPONENT_REGISTRY

from kazoo.exceptions import NoNodeError

# The default size limit for a node in Zookeeper is ~1MiB. However, as this
# also includes the size of the key we can not use all of it for data.
# Because of that we will leave ~47 KiB for the key.
NODE_BYTE_SIZE_LIMIT = 1000000


class RawZKIO(io.RawIOBase):
    def __init__(self, client, path, create=False, makepath=True, version=-1):
        self.client = client
        self.path = path
        self.bytes_read = 0
        self.bytes_written = 0
        self.cumulative_read_time = 0.0
        self.cumulative_write_time = 0.0
        self.znodes_read = 0
        self.znodes_written = 0
        self.create = create
        self.makepath = makepath
        self.version = version
        self.zstat = None

    def readable(self):
        return True

    def writable(self):
        return True

    def truncate(self, size=None):
        # We never truncate unless we're going to write, so make this
        # a noop for the single-znode case.
        pass

    def _getData(self, path):
        start = time.perf_counter()
        data, zstat = self.client.get(path)
        self.cumulative_read_time += time.perf_counter() - start
        self.bytes_read += len(data)
        self.znodes_read += 1
        return data, zstat

    def readall(self):
        data, self.zstat = self._getData(self.path)
        return data

    def write(self, data):
        byte_count = len(data)
        if byte_count > NODE_BYTE_SIZE_LIMIT:
            raise Exception(f"ZK data size too large: {byte_count}")
        start = time.perf_counter()
        if self.create:
            _, self.zstat = self.client.create(
                self.path, data, makepath=self.makepath, include_data=True)
        else:
            self.zstat = self.client.set(self.path, data,
                                         version=self.version)
        self.cumulative_write_time += time.perf_counter() - start
        self.bytes_written += byte_count
        self.znodes_written += 1
        return byte_count


class RawShardIO(RawZKIO):
    def __init__(self, *args, old_format=False, makepath=True, **kw):
        # MODEL_API < 31
        self.old_format = old_format
        self.makepath = makepath
        super().__init__(*args, makepath=makepath, **kw)

    def truncate(self, size=None):
        if size != 0:
            raise ValueError("Can only truncate to 0")
        with suppress(NoNodeError):
            self.client.delete(self.path, recursive=True)
        self.zstat = None

    @property
    def _shards(self):
        start = time.perf_counter()
        ret = self.client.get_children(self.path)
        self.cumulative_read_time += time.perf_counter() - start
        return ret

    def readall_old(self, shard0, shard1):
        # Decompress each shard individually and then recompress them
        # as a unit.
        read_buffer = io.BytesIO()
        read_buffer.write(zlib.decompress(shard0))
        read_buffer.write(zlib.decompress(shard1))
        for shard_count, shard_name in enumerate(sorted(self._shards)):
            if shard_count < 2:
                continue
            shard_path = "/".join((self.path, shard_name))
            data = self._getData(shard_path)[0]
            read_buffer.write(zlib.decompress(data))
        self.zstat = self.client.exists(self.path)
        return zlib.compress(read_buffer.getvalue())

    def readall(self):
        read_buffer = io.BytesIO()
        for shard_count, shard_name in enumerate(sorted(self._shards)):
            shard_path = "/".join((self.path, shard_name))
            data = self._getData(shard_path)[0]
            if shard_count == 1 and data[:2] == b'\x78\x9c':
                # If this is the second shard, and it starts with a
                # zlib header, we're probably reading the old format.
                # Double check that we can decompress it, and if so,
                # switch to reading the old format.
                try:
                    zlib.decompress(data)
                    return self.readall_old(
                        read_buffer.getvalue(),
                        data,
                    )
                except zlib.error:
                    # Perhaps we were wrong about the header
                    pass
            read_buffer.write(self._getData(shard_path)[0])
        self.zstat = self.client.exists(self.path)
        return read_buffer.getvalue()

    def write(self, data):
        # Only write one znode at a time and defer writing the rest to
        # the caller
        data_bytes = bytes(data[0:NODE_BYTE_SIZE_LIMIT])
        read_len = len(data_bytes)
        # MODEL_API < 31
        if self.old_format:
            # We're going to add a header and footer that is several
            # bytes, so we definitely need to reduce the size a little
            # bit, but the old format would end up with a considerable
            # amount of headroom due to compressing after chunking, so
            # lets go ahead and reserve 1k of space.
            new_limit = NODE_BYTE_SIZE_LIMIT - 1024
            data_bytes = data_bytes[0:new_limit]
            # Update our return value to indicate how many bytes we
            # actually read from input.
            read_len = len(data_bytes)
            data_bytes = zlib.compress(data_bytes)
        if not (len(data_bytes) <= NODE_BYTE_SIZE_LIMIT):
            raise RuntimeError("Shard too large")
        start = time.perf_counter()
        # The path we pass to a shard writer is  e.g. '/foo/bar'. Now,
        # for shards the makepath argument should only apply to '/foo'
        # but not the 'bar' subnode as it holds the individual shards
        # and will also be deleted recursively on e.g. a truncate.
        while True:
            try:
                self.client.create(
                    "{}/".format(self.path),
                    data_bytes,
                    sequence=True,
                )
                break
            except NoNodeError:
                if not self.makepath:
                    raise
                self.client.ensure_path(self.path)

        self.cumulative_write_time += time.perf_counter() - start
        self.bytes_written += len(data_bytes)
        self.znodes_written += 1
        if self.zstat is None:
            self.zstat = self.client.exists(self.path)
        return read_len


class BufferedZKWriter(io.BufferedWriter):
    def __init__(self, client, path):
        self.__raw = RawZKIO(client, path)
        super().__init__(self.__raw)

    @property
    def bytes_written(self):
        return self.__raw.bytes_written

    @property
    def cumulative_write_time(self):
        return self.__raw.cumulative_write_time

    @property
    def znodes_written(self):
        return self.__raw.znodes_written

    @property
    def zstat(self):
        return self.__raw.zstat


class BufferedZKReader(io.BufferedReader):
    def __init__(self, client, path):
        self.__raw = RawZKIO(client, path)
        super().__init__(self.__raw)

    @property
    def bytes_read(self):
        return self.__raw.bytes_read

    @property
    def cumulative_read_time(self):
        return self.__raw.cumulative_read_time

    @property
    def znodes_read(self):
        return self.__raw.znodes_read

    @property
    def zstat(self):
        return self.__raw.zstat


class BufferedShardWriter(io.BufferedWriter):
    def __init__(self, client, path, create=False, makepath=True, version=-1):
        self.__old_format = COMPONENT_REGISTRY.model_api < 31
        self.__raw = RawShardIO(client, path, create=create, makepath=makepath,
                                version=version, old_format=self.__old_format)
        super().__init__(self.__raw, NODE_BYTE_SIZE_LIMIT)

    @property
    def bytes_written(self):
        return self.__raw.bytes_written

    @property
    def cumulative_write_time(self):
        return self.__raw.cumulative_write_time

    @property
    def znodes_written(self):
        return self.__raw.znodes_written

    @property
    def zstat(self):
        return self.__raw.zstat

    def write(self, data):
        # MODEL_API < 31
        if self.__old_format and data[:2] == b'\x78\x9c':
            data = zlib.decompress(data)
        return super().write(data)


class BufferedShardReader(io.BufferedReader):
    def __init__(self, client, path):
        self.__raw = RawShardIO(client, path)
        super().__init__(self.__raw, NODE_BYTE_SIZE_LIMIT)

    @property
    def bytes_read(self):
        return self.__raw.bytes_read

    @property
    def cumulative_read_time(self):
        return self.__raw.cumulative_read_time

    @property
    def znodes_read(self):
        return self.__raw.znodes_read

    @property
    def zstat(self):
        return self.__raw.zstat
